# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Codimension main window status bar"""

import os.path
from utils.pixmapcache import getIcon
from plugins.vcssupport.intervaldlg import VCSUpdateIntervalConfigDialog
from .qt import Qt, QPalette, QColor, QMenu, QDialog, QApplication
from .labels import (StatusBarPixmapLabel, StatusBarPathLabel,
                     StatusBarFramedLabel)


class MainWindowStatusBarMixin:

    """Main window status bar mixin"""

    def __init__(self):

        self.__statusBar = None
        self.sbLanguage = None
        self.sbFile = None
        self.sbEol = None
        self.sbPos = None
        self.sbLine = None
        self.sbWritable = None
        self.sbEncoding = None
        self.sbPyflakes = None
        self.sbCC = None
        self.sbVCSStatus = None
        self.sbDebugState = None
        self.__createStatusBar()

    def __createStatusBar(self):
        """Creates status bar"""
        self.__statusBar = self.statusBar()
        self.__statusBar.setSizeGripEnabled(True)

        self.sbVCSStatus = StatusBarPixmapLabel('ignore', self.__statusBar)
        self.__statusBar.addPermanentWidget(self.sbVCSStatus)
        self.sbVCSStatus.setVisible(False)
        self.sbVCSStatus.setContextMenuPolicy(Qt.CustomContextMenu)
        self.sbVCSStatus.customContextMenuRequested.connect(
            self._showVCSLabelContextMenu)

        self.sbDebugState = StatusBarFramedLabel(
            text='Debugger: unknown', callback=None, parent=self.__statusBar)
        dbgPalette = self.sbDebugState.palette()
        dbgPalette.setColor(QPalette.Background, QColor(255, 255, 127))
        self.sbDebugState.setPalette(dbgPalette)
        self.__statusBar.addPermanentWidget(self.sbDebugState)
        self.sbDebugState.setVisible(False)

        self.sbLanguage = StatusBarFramedLabel(parent=self.__statusBar)
        self.__statusBar.addPermanentWidget(self.sbLanguage)

        self.sbEncoding = StatusBarFramedLabel(parent=self.__statusBar)
        self.__statusBar.addPermanentWidget(self.sbEncoding)

        self.sbEol = StatusBarFramedLabel(parent=self.__statusBar)
        self.__statusBar.addPermanentWidget(self.sbEol)

        self.sbWritable = StatusBarFramedLabel(parent=self.__statusBar)
        self.__statusBar.addPermanentWidget(self.sbWritable)

        self.sbPyflakes = StatusBarPixmapLabel('signal', self.__statusBar)
        self.__statusBar.addPermanentWidget(self.sbPyflakes)

        self.sbCC = StatusBarPixmapLabel('signal', self.__statusBar)
        self.__statusBar.addPermanentWidget(self.sbCC)

        self.sbFile = StatusBarPathLabel(
            callback=self._onPathLabelDoubleClick,
            parent=self.__statusBar)
        self.sbFile.setMaximumWidth(512)
        self.sbFile.setMinimumWidth(128)
        self.__statusBar.addPermanentWidget(self.sbFile, True)
        self.sbFile.setContextMenuPolicy(Qt.CustomContextMenu)
        self.sbFile.customContextMenuRequested.connect(
            self._showPathLabelContextMenu)

        self.sbLine = StatusBarFramedLabel(callback=self.copyLine,
                                           parent=self.__statusBar)
        self.sbLine.setMinimumWidth(72)
        self.sbLine.setAlignment(Qt.AlignCenter)
        self.__statusBar.addPermanentWidget(self.sbLine)

        self.sbPos = StatusBarFramedLabel(callback=self.copyPos,
                                          parent=self.__statusBar)
        self.sbPos.setMinimumWidth(72)
        self.sbPos.setAlignment(Qt.AlignCenter)
        self.__statusBar.addPermanentWidget(self.sbPos)

    def copyLine(self):
        """Copies the line number to the buffer"""
        self.__copyLinePos(self.sbLine)

    def copyPos(self):
        """Copies the pos number to the buffer"""
        self.__copyLinePos(self.sbPos)

    @staticmethod
    def __copyLinePos(label):
        """Copies the line/pos label content to the buffer"""
        txt = label.text().strip().lower()
        if not txt.endswith('n/a'):
            txt = txt[txt.find(':') + 1:].strip()
            QApplication.clipboard().setText(txt)

    def showStatusBarMessage(self, msg, timeout=10000):
        """Shows a temporary status bar message, default 10sec"""
        self.__statusBar.showMessage(msg, timeout)

    def clearStatusBarMessage(self):
        """Clears the status bar message in the given slot"""
        self.__statusBar.clearMessage()

    def getCurrentStatusBarMessage(self):
        """Provides the current status bar message"""
        return self.__statusBar.currentMessage()

    def _showVCSLabelContextMenu(self, pos):
        """Triggered when a context menu is requested for a VCS label"""
        contextMenu = QMenu(self)
        contextMenu.addAction(getIcon("vcsintervalmenu.png"),
                              "Configure monitor interval",
                              self.__onVCSMonitorInterval)
        contextMenu.popup(self.sbVCSStatus.mapToGlobal(pos))

    def __onVCSMonitorInterval(self):
        """Runs the VCS monitor interval setting dialog"""
        dlg = VCSUpdateIntervalConfigDialog(
            self.settings['vcsstatusupdateinterval'], self)
        if dlg.exec_() == QDialog.Accepted:
            self.settings['vcsstatusupdateinterval'] = dlg.interval

    def _showPathLabelContextMenu(self, pos):
        """Triggered when a context menu is requested for the path label"""
        contextMenu = QMenu(self)
        contextMenu.addAction(getIcon('copymenu.png'),
                              'Copy full path to clipboard (double click)',
                              self._onPathLabelDoubleClick)
        contextMenu.addSeparator()
        contextMenu.addAction(getIcon(''), 'Copy directory path to clipboard',
                              self._onCopyDirToClipboard)
        contextMenu.addAction(getIcon(''), 'Copy file name to clipboard',
                              self._onCopyFileNameToClipboard)
        contextMenu.popup(self.sbFile.mapToGlobal(pos))

    def _onPathLabelDoubleClick(self):
        """Double click on the status bar path label"""
        txt = self.__getPathLabelFilePath()
        if txt.lower() not in ['', 'n/a']:
            QApplication.clipboard().setText(txt)

    def _onCopyDirToClipboard(self):
        """Copies the dir path of the current file into the clipboard"""
        txt = self.__getPathLabelFilePath()
        if txt.lower() not in ['', 'n/a']:
            try:
                QApplication.clipboard().setText(os.path.dirname(txt) +
                                                 os.path.sep)
            except:
                pass

    def _onCopyFileNameToClipboard(self):
        """Copies the file name of the current file into the clipboard"""
        txt = self.__getPathLabelFilePath()
        if txt.lower() not in ['', 'n/a']:
            try:
                QApplication.clipboard().setText(os.path.basename(txt))
            except:
                pass

    def __getPathLabelFilePath(self):
        """Provides undecorated path label content"""
        txt = str(self.sbFile.getPath())
        if txt.startswith('File: '):
            txt = txt.replace('File: ', '')
        return txt



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2012  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Dialog to edit a single breakpoint"""


from ui.qt import (Qt, QDialog, QDialogButtonBox, QVBoxLayout,
                   QLabel, QGridLayout, QSpinBox, QCheckBox)
from utils.pixmapcache import getIcon
from ui.combobox import CDMComboBox
from .breakpoint import Breakpoint


class BreakpointEditDialog( QDialog ):

    """Dialog with a list of modified but unsaved files implementation"""

    # See utils.run for runParameters
    def __init__(self, bpoint, parent=None):
        QDialog.__init__(self, parent)

        self.__origBpoint = bpoint
        self.setWindowTitle("Edit breakpoint properties")
        self.setWindowIcon(getIcon('bpprops.png'))
        self.__createLayout(bpoint)
        self.__OKButton.setEnabled(False)

        self.__conditionValue.lineEdit().textChanged.connect(self.__changed)
        self.__ignoreValue.valueChanged.connect(self.__changed)
        self.__enabled.stateChanged.connect(self.__changed)
        self.__tempCheckbox.stateChanged.connect(self.__changed)

    def __createLayout(self, bpoint):
        """Creates the dialog layout"""
        self.resize(400, 150)
        self.setSizeGripEnabled(True)

        # Top level layout
        layout = QVBoxLayout(self)

        gridLayout = QGridLayout()
        fileLabel = QLabel("File name:")
        gridLayout.addWidget(fileLabel, 0, 0)
        fileValue = QLabel(bpoint.getAbsoluteFileName())
        gridLayout.addWidget(fileValue, 0, 1)
        lineLabel = QLabel("Line:")
        gridLayout.addWidget(lineLabel, 1, 0)
        lineValue = QLabel(str(bpoint.getLineNumber()))
        gridLayout.addWidget(lineValue, 1, 1)
        conditionLabel = QLabel("Condition:")
        gridLayout.addWidget(conditionLabel, 2, 0)
        self.__conditionValue = CDMComboBox(True)
        self.__conditionValue.lineEdit().setText(bpoint.getCondition())
        gridLayout.addWidget(self.__conditionValue, 2, 1)
        ignoreLabel = QLabel("Ignore count:")
        gridLayout.addWidget(ignoreLabel, 3, 0)
        self.__ignoreValue = QSpinBox()
        self.__ignoreValue.setMinimum(0)
        self.__ignoreValue.setValue(bpoint.getIgnoreCount())
        gridLayout.addWidget(self.__ignoreValue, 3, 1)
        layout.addLayout(gridLayout)

        # Checkboxes part
        self.__tempCheckbox = QCheckBox("&Temporary")
        self.__tempCheckbox.setChecked(bpoint.isTemporary())
        layout.addWidget(self.__tempCheckbox)
        self.__enabled = QCheckBox("&Enabled")
        self.__enabled.setChecked(bpoint.isEnabled())
        layout.addWidget(self.__enabled)

        # Buttons at the bottom
        buttonBox = QDialogButtonBox(self)
        buttonBox.setOrientation(Qt.Horizontal)
        buttonBox.setStandardButtons(QDialogButtonBox.Ok |
                                     QDialogButtonBox.Cancel)
        self.__OKButton = buttonBox.button(QDialogButtonBox.Ok)
        self.__OKButton.setDefault(True)
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.close)
        layout.addWidget(buttonBox)

        self.__conditionValue.setFocus()

    def __changed(self, skipped=None):
        """Triggered when something has been changed"""
        self.__OKButton.setEnabled(True)

    def getData(self):
        """Provides a new instance of a breakpoint"""
        newBPoint = Breakpoint(self.__origBpoint.getAbsoluteFileName(),
                               self.__origBpoint.getLineNumber(),
                               self.__conditionValue.lineEdit().text().strip(),
                               self.__tempCheckbox.isChecked(),
                               self.__enabled.isChecked(),
                               self.__ignoreValue.value())
        return newBPoint




# cml 1 gb id=2 title=Imports
# cml 1 gb id=0 title="Standard imports"
from sys import exit
import os
# cml 1 ge id=0

# cml 1 gb id=1 title="Project imports"
import project_utils
import project_misc
# cml 1 ge id=1
# cml 1 ge id=2

base_path = os.path.curdir()
if project_utils.projectLoaded():
    base_path = project_utils.getBasePath()




# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2014-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Wrapper to run a script with redirected IO"""


import sys
import socket
import traceback
import os.path
import imp
from PyQt5.QtNetwork import QTcpSocket, QAbstractSocket, QHostAddress
from outredir_cdm_dbg import OutStreamRedirector
from cdm_dbg_utils import sendJSONCommand, waitForIDEMessage
from protocol_cdm_dbg import (METHOD_PROC_ID_INFO, METHOD_PROLOGUE_CONTINUE,
                              METHOD_EPILOGUE_EXIT, METHOD_EPILOGUE_EXIT_CODE,
                              METHOD_STDIN)


CLIENT_DEBUG = True

WAIT_CONTINUE_TIMEOUT = 5       # in seconds
WAIT_EXIT_COMMAND_TIMEOUT = 5   # in seconds
RUN_WRAPPER = None
RUN_CLIENT_ORIG_INPUT = None


def runClientInput(prompt="", echo=True):
    """Replacement for the standard raw_input builtin"""
    if RUN_WRAPPER is None or not RUN_WRAPPER.redirected():
        return RUN_CLIENT_ORIG_INPUT(prompt)
    return RUN_WRAPPER.input(prompt, echo)


# Use our own input().
try:
    RUN_CLIENT_ORIG_INPUT = __builtins__.__dict__['input']
    __builtins__.__dict__['input'] = runClientInput
except (AttributeError, KeyError):
    import __main__
    RUN_CLIENT_ORIG_INPUT = __main__.__builtins__.__dict__['input']
    __main__.__builtins__.__dict__['input'] = runClientInput


class RedirectedIORunWrapper():

    """Wrapper to run a script with redirected IO"""

    def __init__(self):
        self.__socket = None
        self.__redirected = False
        self.__procuuid = None

    def redirected(self):
        """True if streams are redirected"""
        return self.__redirected

    def main(self):
        """Run wrapper driver"""
        if '--' not in sys.argv:
            print("Unexpected arguments", file=sys.stderr)
            return 1

        self.__procuuid, host, port, args = self.parseArgs()
        if self.__procuuid is None or host is None or port is None:
            print("Not enough arguments", file=sys.stderr)
            return 1

        remoteAddress = self.resolveHost(host)
        self.connect(remoteAddress, port)
        sendJSONCommand(self.__socket, METHOD_PROC_ID_INFO,
                        self.__procuuid, None)

        try:
            waitForIDEMessage(self.__socket, METHOD_PROLOGUE_CONTINUE,
                              WAIT_CONTINUE_TIMEOUT)
        except Exception as exc:
            print(str(exc), file=sys.stderr)
            return 1

        # Setup redirections
        stdoutOld = sys.stdout
        stderrOld = sys.stderr
        sys.stdout = OutStreamRedirector(self.__socket, True, self.__procuuid)
        sys.stderr = OutStreamRedirector(self.__socket, False, self.__procuuid)
        self.__redirected = True

        # Run the script
        retCode = 0
        try:
            self.__runScript(args)
        except SystemExit as exc:
            if CLIENT_DEBUG:
                print(traceback.format_exc(), file=sys.__stderr__)
            if exc.code is None:
                retCode = 0
            elif isinstance(exc.code, int):
                retCode = exc.code
            else:
                retCode = 1
                print(str(exc.code), file=sys.stderr)
        except KeyboardInterrupt as exc:
            if CLIENT_DEBUG:
                print(traceback.format_exc(), file=sys.__stderr__)
            retCode = 1
            print(traceback.format_exc(), file=sys.stderr)
        except Exception as exc:
            if CLIENT_DEBUG:
                print(traceback.format_exc(), file=sys.__stderr__)
            retCode = 1
            print(traceback.format_exc(), file=sys.stderr)
        except:
            if CLIENT_DEBUG:
                print(traceback.format_exc(), file=sys.__stderr__)
            retCode = 1
            print(traceback.format_exc(), file=sys.stderr)

        sys.stderr = stderrOld
        sys.stdout = stdoutOld
        self.__redirected = False

        # Send the return code back
        try:
            sendJSONCommand(self.__socket, METHOD_EPILOGUE_EXIT_CODE,
                            self.__procuuid, {'exitCode': retCode})
        except Exception as exc:
            print(str(exc), file=sys.stderr)
            self.close()
            return 1

        self.close()
        return 0

    def connect(self, remoteAddress, port):
        """Establishes a connection with the IDE"""
        self.__socket = QTcpSocket()
        if remoteAddress is None:
            self.__socket.connectToHost(QHostAddress.LocalHost, port)
        else:
            self.__socket.connectToHost(remoteAddress, port)
        if not self.__socket.waitForConnected(1000):
            raise Exception('Cannot connect to the IDE')
        self.__socket.setSocketOption(QAbstractSocket.KeepAliveOption, 1)
        self.__socket.setSocketOption(QAbstractSocket.LowDelayOption, 1)
        self.__socket.disconnected.connect(self.__onDisconnected)

    def __onDisconnected(self):
        """IDE dropped the socket"""
        sys.exit(0)

    def __runScript(self, arguments):
        """Runs the python script"""
        try:
            # In Py 2.x, the builtins were in __builtin__
            builtins = sys.modules['__builtin__']
        except KeyError:
            # In Py 3.x, they're in builtins
            builtins = sys.modules['builtins']

        fileName = arguments[0]

        oldMainMod = sys.modules['__main__']
        mainMod = imp.new_module('__main__')
        sys.modules['__main__'] = mainMod
        mainMod.__file__ = fileName
        mainMod.__builtins__ = builtins

        # Set sys.argv properly.
        oldArgv = sys.argv
        sys.argv = arguments

        # Without this imports of what is located at the script directory do
        # not work
        normCWD = os.path.normpath(os.getcwd())
        normScriptDir = os.path.normpath(os.path.dirname(fileName))
        sys.path.insert(0, os.getcwd())
        if normCWD != normScriptDir:
            sys.path.insert(0, os.path.dirname(fileName))

        with open(fileName, 'rb') as fileToRead:
            source = fileToRead.read()

        # We have the source.  `compile` still needs the last line to be clean,
        # so make sure it is, then compile a code object from it.
        if not source or source[-1] != b'\n':
            source += b'\n'

        code = compile(source, fileName, "exec")
        exec(code, mainMod.__dict__)

        # Restore the old __main__
        sys.modules['__main__'] = oldMainMod

        # Restore the old argv and path
        sys.argv = oldArgv

    def close(self):
        """Closes the connection if so"""
        try:
            if self.__socket:
                # Wait exit needed because otherwise IDE may get socket
                # disconnected before it has a chance to read the script
                # exit code. Wait for the explicit command to exit guarantees
                # that all the data will be received.
                waitForIDEMessage(self.__socket, METHOD_EPILOGUE_EXIT,
                                  WAIT_EXIT_COMMAND_TIMEOUT)
        finally:
            if self.__socket:
                self.__socket.close()
                self.__socket = None

    def input(self, prompt, echo):
        """Implements 'input' using the redirected input"""
        sendJSONCommand(self.__socket, METHOD_STDIN, self.__procuuid,
                        {'prompt': prompt, 'echo': echo})
        params = waitForIDEMessage(self.__socket, METHOD_STDIN,
                                   60 * 60 * 24 * 7)
        return params['input']

    @staticmethod
    def resolveHost(host):
        """Resolves a hostname to an IP address"""
        try:
            host, _ = host.split("@@")
            family = socket.AF_INET6
        except ValueError:
            # version = 'v4'
            family = socket.AF_INET
        return socket.getaddrinfo(host, None, family,
                                  socket.SOCK_STREAM)[0][4][0]

    @staticmethod
    def parseArgs():
        """Parses the arguments"""
        host = None
        port = None
        procuuid = None
        args = sys.argv[1:]

        while args[0]:
            if args[0] in ['--host']:
                host = args[1]
                del args[0]
                del args[0]
            elif args[0] in ['--port']:
                port = int(args[1])
                del args[0]
                del args[0]
            elif args[0] in ['--procuuid']:
                procuuid = args[1]
                del args[0]
                del args[0]
            elif args[0] == '--':
                del args[0]
                break

        return procuuid, host, port, args


if __name__ == "__main__":
    RUN_WRAPPER = RedirectedIORunWrapper()
    sys.exit(RUN_WRAPPER.main())



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""QComboBox extension:
   - allows editing
   - limits the number of saved entries
   - does not allow duplications
   - disables auto completion
   - inserts items at top
"""

from .qt import Qt, pyqtSignal, QComboBox


class EnterSensitiveComboBox(QComboBox):

    """Combo box which emits 'enterClicked' signal"""

    enterClicked = pyqtSignal()

    def __init__(self, parent=None):
        QComboBox.__init__(self, parent)

    def keyReleaseEvent(self, event):
        """Triggered when a key is released"""
        QComboBox.keyReleaseEvent(self, event)
        key = event.key()
        if key == Qt.Key_Enter or key == Qt.Key_Return:
            self.enterClicked.emit()


class CDMComboBox(QComboBox):

    """QComboBox minor extension"""

    enterClicked = pyqtSignal()
    itemAdded = pyqtSignal()

    itemsLimit = 32

    def __init__(self, insertOnEnter=True, parent=None):
        QComboBox.__init__(self, parent)
        self.setEditable(True)
        self.setCompleter(None)
        self.setDuplicatesEnabled(False)
        self.setInsertPolicy(QComboBox.NoInsert)
        self.insert = insertOnEnter

    def addItem(self, text):
        """Adds an item to the top"""
        text = str(text).strip()
        if text != "":
            if self.findText(text, Qt.MatchFixedString) == -1:
                self.insertItem(0, text)
                self.__enforceLimit()
        self.itemAdded.emit()

    def getItems(self):
        """Provides the list of items (text only)"""
        items = []
        for index in range(self.count()):
            items.append(str(self.itemText(index)))
        return items

    def keyReleaseEvent(self, event):
        """Triggered when a key is released"""
        QComboBox.keyReleaseEvent(self, event)
        if self.insert:
            key = event.key()
            if key == Qt.Key_Enter or key == Qt.Key_Return:
                self.addItem(self.lineEdit().text())
                self.__enforceLimit()
                self.enterClicked.emit()

    def __enforceLimit(self):
        """checks the number of memorized items"""
        # Check that the number of items is not exceeded
        while self.count() > self.itemsLimit:
            self.removeItem(self.count() - 1)



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2018  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Virtual canvas items to handle groups (opened/collapsed)"""

# pylint: disable=C0305
# pylint: disable=R0902
# pylint: disable=R0913

from ui.qt import (Qt, QPen, QBrush, QGraphicsRectItem, QGraphicsItem, QFrame,
                   QLabel, QVBoxLayout, QPalette, QApplication)
from utils.globals import GlobalData
from .cellelement import CellElement
from .auxitems import Connector, SpacerCell
from .colormixin import ColorMixin
from .textmixin import TextMixin


class HGroupSpacerCell(SpacerCell):

    """Represents a horizontal spacer cell used to shift items due to groups"""

    def __init__(self, ref, canvas, x, y):
        SpacerCell.__init__(self, ref, canvas, x, y, width=0, height=0)
        self.kind = CellElement.H_GROUP_SPACER

        # Number of spacers to be inserted
        self.count = 0

    def render(self):
        """Renders the cell"""
        self.width = self.count * 2 * self.canvas.settings.openGroupHSpacer
        self.height = 0
        self.minWidth = self.width
        self.minHeight = self.height
        return (self.width, self.height)


class GroupItemBase(CellElement, TextMixin):

    """Common functionality for the group items"""

    def __init__(self, groupBeginCMLRef, ref, canvas, x, y):
        CellElement.__init__(self, ref, canvas, x, y)
        TextMixin.__init__(self)

        self.nestedRefs = []

        self.groupBeginCMLRef = groupBeginCMLRef
        self.groupEndCMLRef = None

        # True if the previous item is terminal, i.e no connector needed
        self.isTerminal = False

    def getGroupId(self):
        """Provides the group ID"""
        return self.groupBeginCMLRef.id

    def getTitle(self):
        """Convenience for the UI"""
        return self.groupBeginCMLRef.getTitle()

    def getLineRange(self):
        """Provides the line range"""
        return self.groupBeginCMLRef.ref.getLineRange()

    def getAbsPosRange(self):
        """Provides the absolute position range"""
        return self.groupBeginCMLRef.ref.getAbsPosRange()

    def getSelectTooltip(self):
        """Provides the tooltip"""
        return 'Group at ' + CellElement.getLinesSuffix(self.getLineRange())



class EmptyGroup(GroupItemBase, ColorMixin, QGraphicsRectItem):

    """Represents an empty group"""

    N_BACK_RECT = 2

    def __init__(self, ref, groupBeginCMLRef, canvas, x, y):
        GroupItemBase.__init__(self, groupBeginCMLRef, ref, canvas, x, y)
        ColorMixin.__init__(self, None, canvas.settings.emptyGroupBGColor,
                            canvas.settings.emptyGroupFGColor,
                            canvas.settings.emptyGroupBorderColor,
                            colorSpec=groupBeginCMLRef)
        QGraphicsRectItem.__init__(self, canvas.scopeRectangle)
        self.kind = CellElement.EMPTY_GROUP

        self.connector = None

        # To make double click delivered
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)

    def render(self):
        """Renders the cell"""
        settings = self.canvas.settings
        self.setupText(self, customText=self.getTitle(),
                       customReplacement='')

        vPadding = 2 * (settings.vCellPadding + settings.vTextPadding) + \
                   self.N_BACK_RECT * settings.emptyGroupYShift
        self.minHeight = self.textRect.height() + vPadding
        hPadding = 2 * (settings.hCellPadding + settings.hTextPadding) + \
                   self.N_BACK_RECT * settings.emptyGroupXShift
        self.minWidth = max(self.textRect.width() + hPadding,
                            settings.minWidth)
        self.height = self.minHeight
        self.width = self.minWidth
        return (self.width, self.height)

    def draw(self, scene, baseX, baseY):
        """Draws the cell"""
        self.baseX = baseX
        self.baseY = baseY

        # Add the connector as a separate scene item to make the selection
        # working properly
        settings = self.canvas.settings
        self.connector = Connector(self.canvas, baseX + settings.mainLine,
                                   baseY,
                                   baseX + settings.mainLine,
                                   baseY + self.height)
        scene.addItem(self.connector)

        # Setting the rectangle is important for the selection and for
        # redrawing. Thus the selection pen with must be considered too.
        penWidth = settings.selectPenWidth - 1
        self.setRect(baseX + settings.hCellPadding - penWidth,
                     baseY + settings.vCellPadding - penWidth,
                     self.minWidth - 2 * settings.hCellPadding + 2 * penWidth,
                     self.minHeight - 2 * settings.vCellPadding + 2 * penWidth)
        scene.addItem(self)

    def paint(self, painter, option, widget):
        """Draws the collapsed group"""
        del option      # unused argument
        del widget      # unused argument

        settings = self.canvas.settings
        pen = self.getPainterPen(self.isSelected(), self.borderColor)
        if not self.isSelected():
            pen.setStyle(Qt.DotLine)
        painter.setPen(pen)
        painter.setBrush(QBrush(self.bgColor))

        # Outer rectangle
        rectWidth = self.minWidth - 2 * settings.hCellPadding - \
                    self.N_BACK_RECT * settings.emptyGroupXShift
        rectHeight = self.minHeight - 2 * settings.vCellPadding - \
                     self.N_BACK_RECT * settings.emptyGroupYShift

        for rectNum in range(self.N_BACK_RECT, -1, -1):
            xPos = self.baseX + settings.hCellPadding + \
                   rectNum * settings.emptyGroupXShift
            yPos = self.baseY + settings.vCellPadding + \
                   (self.N_BACK_RECT - rectNum) * settings.emptyGroupYShift
            painter.drawRect(xPos, yPos, rectWidth, rectHeight)

        # Draw the text in the rectangle
        pen = QPen(self.fgColor)
        painter.setFont(settings.monoFont)
        painter.setPen(pen)

        textWidth = self.textRect.width() + 2 * settings.hTextPadding
        textShift = (rectWidth - textWidth) / 2
        painter.drawText(
            self.baseX + settings.hCellPadding +
            settings.hTextPadding +
            textShift,
            self.baseY + settings.vCellPadding + settings.vTextPadding +
            self.N_BACK_RECT * settings.emptyGroupYShift,
            self.textRect.width(), self.textRect.height(),
            Qt.AlignLeft, self.text)


class OpenedGroupBegin(GroupItemBase, ColorMixin, QGraphicsRectItem):

    """Represents beginning af a group which can be collapsed"""

    def __init__(self, ref, groupBeginCMLRef, canvas, x, y):
        GroupItemBase.__init__(self, groupBeginCMLRef, ref, canvas, x, y)
        ColorMixin.__init__(self, None, canvas.settings.openGroupBGColor,
                            canvas.settings.openGroupFGColor,
                            canvas.settings.openGroupBorderColor,
                            colorSpec=groupBeginCMLRef)
        QGraphicsRectItem.__init__(self, canvas.scopeRectangle)
        self.kind = CellElement.OPENED_GROUP_BEGIN
        self.connector = None
        self.topLeftControl = None
        self.highlight = False

        # These two items are filled when rendering is finished for all the
        # items in the group
        self.groupWidth = None
        self.groupHeight = None

        self.groupEndRow = None
        self.groupEndColumn = None

        self.selfAndDeeperNestLevel = None
        self.selfMaxNestLevel = None    # Used in vcanvas.py

        # To make double click delivered
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)

    def setHighlight(self, newValue):
        """Changes the highlight status of the group"""
        if self.highlight != newValue:
            self.highlight = newValue
            if not self.isSelected():
                self.update()

    def render(self):
        """Renders the cell"""
        self.topLeftControl = GroupCornerControl(self)
        self.width = self.canvas.settings.openGroupHSpacer * 2
        self.height = self.canvas.settings.openGroupVSpacer * 2
        self.minWidth = self.width
        self.minHeight = self.height
        return (self.width, self.height)

    def draw(self, scene, baseX, baseY):
        """Draws the cell"""
        self.baseX = baseX
        self.baseY = baseY

        settings = self.canvas.settings

        # Setting the rectangle is important for the selection and for
        # redrawing. Thus the selection pen with must be considered too.
        penWidth = settings.selectPenWidth - 1

        groupWidth = self.groupWidth + 2 * settings.openGroupHSpacer

        self.setRect(baseX - penWidth + settings.openGroupHSpacer,
                     baseY - penWidth + settings.openGroupVSpacer,
                     groupWidth + 2 * penWidth,
                     self.groupHeight +
                     2 * (penWidth + settings.openGroupVSpacer))
        scene.addItem(self)

        # Add the connector as a separate scene item to make the selection
        # working properly. The connector must be added after a group,
        # otherwise a half of it is hidden by the group.
        if not self.isTerminal:
            xPos = baseX + settings.mainLine
            xPos += \
                self.selfAndDeeperNestLevel * (2 * settings.openGroupHSpacer)
            self.connector = Connector(self.canvas, xPos, baseY, xPos,
                                       baseY + settings.openGroupVSpacer * 2)
            scene.addItem(self.connector)

        # Top left corner control
        self.topLeftControl.moveTo(baseX, baseY)
        scene.addItem(self.topLeftControl)

    def paint(self, painter, option, widget):
        """Draws the collapsed group"""
        del option      # unused argument
        del widget      # unused argument

        settings = self.canvas.settings
        pen = self.getPainterPen(self.isSelected(), self.borderColor)
        if not self.highlight and not self.isSelected():
            pen.setStyle(Qt.DotLine)
        painter.setPen(pen)
        painter.setBrush(QBrush(self.bgColor))

        # Group rectangle
        fullWidth = self.groupWidth + 2 * settings.openGroupHSpacer
        fullHeight = self.groupHeight + 2 * settings.openGroupVSpacer
        painter.drawRoundedRect(self.baseX + settings.openGroupHSpacer,
                                self.baseY + settings.openGroupVSpacer,
                                fullWidth, fullHeight,
                                settings.openGroupVSpacer,
                                settings.openGroupVSpacer)


class OpenedGroupEnd(GroupItemBase):

    """Represents the end af a group which can be collapsed"""

    def __init__(self, ref, groupBeginCMLRef, canvas, x, y):
        GroupItemBase.__init__(self, groupBeginCMLRef, ref, canvas, x, y)
        self.kind = CellElement.OPENED_GROUP_END

        self.groupBeginRow = None
        self.groupBeginColumn = None
        self.connector = None

        self.selfAndDeeperNestLevel = None

    def render(self):
        """Renders the cell"""
        self.width = 0
        self.height = self.canvas.settings.openGroupVSpacer * 2
        self.minWidth = self.width
        self.minHeight = self.height
        return (self.width, self.height)

    def draw(self, scene, baseX, baseY):
        """Draws the cell"""
        self.baseX = baseX
        self.baseY = baseY

        if self.isTerminal:
            return

        # Add the connector as a separate scene item to make the selection
        # working properly
        settings = self.canvas.settings
        xPos = baseX + settings.mainLine
        xPos += self.selfAndDeeperNestLevel * (2 * settings.openGroupHSpacer)
        self.connector = Connector(self.canvas,
                                   xPos,
                                   baseY,
                                   xPos,
                                   baseY + settings.openGroupVSpacer * 2)
        scene.addItem(self.connector)



class CollapsedGroup(GroupItemBase, ColorMixin, QGraphicsRectItem):

    """Represents a collapsed group"""

    N_BACK_RECT = 2

    def __init__(self, ref, groupBeginCMLRef, canvas, x, y):
        GroupItemBase.__init__(self, groupBeginCMLRef, ref, canvas, x, y)
        ColorMixin.__init__(self, None, canvas.settings.collapsedGroupBGColor,
                            canvas.settings.collapsedGroupFGColor,
                            canvas.settings.collapsedGroupBorderColor,
                            colorSpec=groupBeginCMLRef)
        QGraphicsRectItem.__init__(self, canvas.scopeRectangle)
        self.kind = CellElement.COLLAPSED_GROUP
        self.connector = None

        # To make double click delivered
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)

    def render(self):
        """Renders the cell"""
        settings = self.canvas.settings
        self.setupText(self, customText=self.getTitle(),
                       customReplacement='')

        vPadding = 2 * (settings.vCellPadding + settings.vTextPadding) + \
                   self.N_BACK_RECT * settings.collapsedGroupYShift
        self.minHeight = self.textRect.height() + vPadding
        hPadding = 2 * (settings.hCellPadding + settings.hTextPadding) + \
                   self.N_BACK_RECT * settings.collapsedGroupXShift
        self.minWidth = max(self.textRect.width() + hPadding,
                            settings.minWidth)
        self.height = self.minHeight
        self.width = self.minWidth
        return (self.width, self.height)

    def draw(self, scene, baseX, baseY):
        """Draws the cell"""
        self.baseX = baseX
        self.baseY = baseY

        # Add the connector as a separate scene item to make the selection
        # working properly
        settings = self.canvas.settings
        self.connector = Connector(self.canvas, baseX + settings.mainLine,
                                   baseY,
                                   baseX + settings.mainLine,
                                   baseY + self.height)
        scene.addItem(self.connector)

        # Setting the rectangle is important for the selection and for
        # redrawing. Thus the selection pen with must be considered too.
        penWidth = settings.selectPenWidth - 1
        self.setRect(baseX + settings.hCellPadding - penWidth,
                     baseY + settings.vCellPadding - penWidth,
                     self.minWidth - 2 * settings.hCellPadding + 2 * penWidth,
                     self.minHeight - 2 * settings.vCellPadding + 2 * penWidth)
        scene.addItem(self)

    def paint(self, painter, option, widget):
        """Draws the collapsed group"""
        del option      # unused argument
        del widget      # unused argument

        settings = self.canvas.settings
        painter.setPen(self.getPainterPen(self.isSelected(), self.borderColor))
        painter.setBrush(QBrush(self.bgColor))

        # Outer rectangle
        rectWidth = self.minWidth - 2 * settings.hCellPadding - \
                    self.N_BACK_RECT * settings.collapsedGroupXShift
        rectHeight = self.minHeight - 2 * settings.vCellPadding - \
                     self.N_BACK_RECT * settings.collapsedGroupYShift

        for rectNum in range(self.N_BACK_RECT, -1, -1):
            xPos = self.baseX + settings.hCellPadding + \
                   rectNum * settings.collapsedGroupXShift
            yPos = self.baseY + settings.vCellPadding + \
                   (self.N_BACK_RECT - rectNum) * settings.collapsedGroupYShift
            painter.drawRect(xPos, yPos, rectWidth, rectHeight)

        # Draw the text in the rectangle
        pen = QPen(self.fgColor)
        painter.setFont(settings.monoFont)
        painter.setPen(pen)

        textWidth = self.textRect.width() + 2 * settings.hTextPadding
        textShift = (rectWidth - textWidth) / 2
        painter.drawText(
            self.baseX + settings.hCellPadding +
            settings.hTextPadding +
            textShift,
            self.baseY + settings.vCellPadding + settings.vTextPadding +
            self.N_BACK_RECT * settings.collapsedGroupYShift,
            self.textRect.width(), self.textRect.height(),
            Qt.AlignLeft, self.text)



class GroupCornerControl(CellElement, QGraphicsRectItem):

    """Expanded group top left corner control"""

    def __init__(self, ref):
        CellElement.__init__(self, ref, ref.canvas, None, None)
        QGraphicsRectItem.__init__(self)
        self.kind = CellElement.GROUP_CORNER_CONROL

        settings = self.canvas.settings
        self.__width = settings.openGroupHSpacer * 2 - 1
        self.__height = settings.openGroupVSpacer * 2 - 1

        self.setAcceptHoverEvents(True)

    def moveTo(self, xPos, yPos):
        """Moves to the specified position"""
        settings = self.canvas.settings

        # This is a mistery. I do not understand why I need to divide by 2.0
        # however this works. I tried various combinations of initialization,
        # setting the position and mapping. Nothing works but ../2.0. Sick!
        self.setPos((float(xPos) + settings.openGroupHSpacer - 1) / 2.0,
                    (float(yPos) + settings.openGroupVSpacer - 1) / 2.0)
        self.setRect(self.x(), self.y(),
                     self.__width, self.__height)

    def paint(self, painter, option, widget):
        """Paints the control"""
        del option
        del widget

        settings = self.canvas.settings

        pen = QPen(settings.openGroupControlBorderColor)
        pen.setStyle(Qt.SolidLine)
        pen.setWidth(settings.openGroupControlLineWidth)
        painter.setPen(pen)
        painter.setBrush(QBrush(settings.openGroupControlBGColor))

        painter.drawRoundedRect(self.x(), self.y(),
                                self.__width, self.__height,
                                1, 1)

    def hoverEnterEvent(self, event):
        """Handles the mouse in event"""
        del event
        self.ref.setHighlight(True)
        if self.ref.getTitle():
            groupTitlePopup.setTitleForGroup(self.ref)
            groupTitlePopup.show(self)

    def hoverLeaveEvent(self, event):
        """Handles the mouse out event"""
        del event
        self.ref.setHighlight(False)
        groupTitlePopup.hide()

    def getProxiedItem(self):
        """Provides the real item for a proxy one"""
        return self.ref



class GroupTitlePopup(QFrame):

    """Frameless panel to show the group title"""

    def __init__(self, parent):
        QFrame.__init__(self, parent)

        self.setWindowFlags(Qt.SplashScreen |
                            Qt.WindowStaysOnTopHint |
                            Qt.X11BypassWindowManagerHint)

        self.setFrameShape(QFrame.StyledPanel)
        self.setLineWidth(1)

        self.__titleLabel = None
        self.__createLayout()

    def __createLayout(self):
        """Creates the widget layout"""
        verticalLayout = QVBoxLayout(self)
        verticalLayout.setContentsMargins(0, 0, 0, 0)

        self.__titleLabel = QLabel()
        self.__titleLabel.setAutoFillBackground(True)
        self.__titleLabel.setFrameShape(QFrame.StyledPanel)
        self.__titleLabel.setStyleSheet('padding: 2px')
        verticalLayout.addWidget(self.__titleLabel)

    def setTitleForGroup(self, group):
        """Sets the title of the group"""
        self.__titleLabel.setFont(group.canvas.settings.monoFont)
        self.__titleLabel.setText(group.getTitle())

    def resize(self, controlItem):
        """Moves the popup to the proper position"""
        # calculate the positions above the group
        # Taken from here:
        # https://stackoverflow.com/questions/9871749/find-screen-position-of-a-qgraphicsitem
        scene = controlItem.ref.scene()
        view = scene.views()[0]
        sceneP = controlItem.mapToScene(controlItem.boundingRect().topLeft())
        viewP = view.mapFromScene(sceneP)
        pos = view.viewport().mapToGlobal(viewP)

        self.move(pos.x(), pos.y() - self.height() - 2)
        QApplication.processEvents()

    def show(self, controlItem):
        """Shows the title above the group control"""
        # Use the palette from the group
        bgColor, fgColor, _ = controlItem.ref.getColors()
        palette = self.__titleLabel.palette()
        palette.setColor(QPalette.Background, bgColor)
        palette.setColor(QPalette.Foreground, fgColor)
        self.__titleLabel.setPalette(palette)

        # That's a trick: resizing works correctly only if the item is shown
        # So move it outside of the screen, show it so it is invisible and then
        # resize and move to the proper position
        screenHeight = GlobalData().screenHeight
        self.move(0, screenHeight + 128)
        QFrame.show(self)
        QApplication.processEvents()
        self.resize(controlItem)

# One instance use for all
groupTitlePopup = GroupTitlePopup(None)




# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2011-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""The diff viewer implementation"""

from utils.pixmapcache import getIcon
from utils.globals import GlobalData
from .qt import (Qt, QSize, QHBoxLayout, QWidget, QAction, QToolBar,
                 QSizePolicy, QVBoxLayout)
from .htmltabwidget import HTMLTabWidget


class DiffViewer(QWidget):

    """The diff viewer widget at the bottom"""

    NODIFF = None

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)

        self.viewer = None
        self.__clearButton = None
        self.__sendUpButton = None
        self.__createLayout()
        self.__isEmpty = True
        self.__tooltip = ""
        self.__inClear = False

        paperColor = GlobalData().skin['nolexerPaper'].name()
        NODIFF = '<html><body bgcolor="' + paperColor + '"></body></html>'
        self.viewer.setHTML(self.NODIFF)
        self.__updateToolbarButtons()

    def __createLayout(self):
        """Helper to create the viewer layout"""
        self.viewer = HTMLTabWidget()

        # Buttons
        self.__sendUpButton = QAction(getIcon('senddiffup.png'),
                                      'Send to Main Editing Area', self)
        self.__sendUpButton.triggered.connect(self.__sendUp)
        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.__clearButton = QAction(getIcon('trash.png'),
                                     'Clear Generated Diff', self)
        self.__clearButton.triggered.connect(self.__clear)

        # Toolbar
        self.toolbar = QToolBar()
        self.toolbar.setOrientation(Qt.Vertical)
        self.toolbar.setMovable(False)
        self.toolbar.setAllowedAreas(Qt.LeftToolBarArea)
        self.toolbar.setIconSize(QSize(16, 16))
        self.toolbar.setFixedWidth(28)
        self.toolbar.setContentsMargins(0, 0, 0, 0)
        self.toolbar.addAction(self.__sendUpButton)
        self.toolbar.addWidget(spacer)
        self.toolbar.addAction(self.__clearButton)

        verticalLayout = QVBoxLayout()
        verticalLayout.setContentsMargins(2, 2, 2, 2)
        verticalLayout.setSpacing(2)
        verticalLayout.addWidget(self.viewer)

        # layout
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        layout.addWidget(self.toolbar)
        layout.addLayout(verticalLayout)

        self.setLayout(layout)

    def setHTML(self, content, tooltip):
        """Shows the given content"""
        if self.__inClear:
            self.viewer.setHTML(content)
            self.viewer.onTextZoomChanged()
            return

        if content == '' or content is None:
            self.__clear()
        else:
            self.viewer.setHTML(content)
            self.viewer.onTextZoomChanged()
            self.__isEmpty = False
            self.__updateToolbarButtons()
            self.__tooltip = tooltip

    def onTextZoomChanged(self):
        """Triggered when a text zoom is changed"""
        self.viewer.onTextZoomChanged()

    def __sendUp(self):
        """Triggered when the content should be sent to the main editor area"""
        if not self.__isEmpty:
            GlobalData().mainWindow.showDiffInMainArea(self.viewer.getHTML(),
                                                       self.__tooltip)

    def __clear(self):
        """Triggered when the content should be cleared"""
        self.__inClear = True
        # Dirty hack - reset the tooltip
        GlobalData().mainWindow.showDiff("", "No diff shown")
        self.viewer.setHTML(DiffViewer.NODIFF)
        self.__inClear = False

        self.__isEmpty = True
        self.__tooltip = ""
        self.__updateToolbarButtons()

    def __updateToolbarButtons(self):
        """Contextually updates toolbar buttons"""
        self.__sendUpButton.setEnabled(not self.__isEmpty)
        self.__clearButton.setEnabled(not self.__isEmpty)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2018  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Codimension main Python script.

It performs necessery initialization and starts the Qt main loop.
"""

# pylint: disable=W0702
# pylint: disable=W0703
# pylint: disable=C0415
# pylint: disable=C0413

import sys
import os
import os.path
import gc
import traceback
import logging
import datetime
from optparse import OptionParser

# Workaround if link is used
sys.argv[0] = os.path.realpath(__file__)

# Make it possible to import from the subdirectories
originalSysPath = sys.path
srcDir = os.path.dirname(sys.argv[0])
if srcDir not in sys.path:
    sys.path.insert(0, srcDir)

# Use the python parser in the qutepart component
os.environ['QPART_CPARSER'] = 'N'

from ui.qt import QTimer, QMessageBox
from ui.application import CodimensionApplication
from ui.splashscreen import SplashScreen
from utils.project import CodimensionProject
from utils.skin import Skin, populateSampleSkin
from utils.config import DEFAULT_ENCODING
from utils.settings import Settings, SETTINGS_DIR
from utils.globals import GlobalData

# Need for the correct import resolution
GlobalData().originalSysPath = originalSysPath


try:
    import cdmverspec
    VER = cdmverspec.version
except:
    VER = '0.0.0'


class CodimensionUILauncher:

    """Wrapper class to launch Codimension"""

    def __init__(self):
        self.rootLoggingHandlers = []

        # In case of windows XServers (at least X-Win32) early usage of
        # redirected logging.something(...) leads to a silent crash. It does
        # not happen in a pure Linux environment though. So some warning
        # messages are delayed till the main application loop has started.
        self.__delayedWarnings = []

        self.__args = []
        self.__options = None
        self.__projectFile = None
        self.__splash = None

    def setupLogging(self):
        """Configures the logging module"""
        if self.__options.debug:
            logLevel = logging.DEBUG
        else:
            logLevel = logging.INFO

        # Default output stream is stderr
        logging.basicConfig(level=logLevel,
                            format='%(levelname) -10s %(asctime)s %(message)s')

        # Memorize the root logging handlers
        self.rootLoggingHandlers = logging.root.handlers

    def isOpenable(self, fName):
        """Checks that the file can be opened"""
        # I cannot import it at the top because the fileutils want
        # to use the pixmap cache which needs the application to be
        # created, so the import is deferred
        from utils.fileutils import isFileOpenable

        if not os.path.exists(fName):
            self.__delayedWarnings.append('File does not exist: ' + fName)
            return False
        if not os.path.isfile(fName):
            self.__delayedWarnings.append('The ' + fName + ' is not a file')
            return False
        if not isFileOpenable(fName):
            self.__delayedWarnings.append('The file ' + fName +
                                          ' could not be opened in Codimension')
            return False
        return True

    def processCommandLineArgs(self):
        """Checks what is in the command line"""
        # I cannot import it at the top because the fileutils want
        # to use the pixmap cache which needs the application to be
        # created, so the import is deferred
        from utils.fileutils import getFileProperties, isCDMProjectMime

        if self.__args:
            totalArgs = len(self.__args)
            goodArgs = []
            for fName in self.__args:
                if self.isOpenable(fName):
                    goodArgs.append(fName)
            self.__args = goodArgs

            if totalArgs == 1 and len(self.__args) == 1:
                mime, _, _ = getFileProperties(self.__args[0])
                if isCDMProjectMime(mime):
                    return self.__args[0]

            # Check that the is no project file because it can only come alone
            for fName in self.__args:
                mime, _, _ = getFileProperties(fName)
                if isCDMProjectMime(mime):
                    raise Exception('A Codimension project file (' +
                                    fName + ') must not come '
                                    'together with other files')
        return None

    def main(self):
        """The codimension driver"""
        usageMessage = 'Usage: %prog [options] [project file | python files]'
        parser = OptionParser(usage=usageMessage, version='%prog ' + VER)

        parser.add_option(
            '--debug', action='store_true', dest='debug',
            default=False,
            help='switch on debug and info messages (default: Off)')
        parser.add_option(
            '--clean-start', action='store_true', dest='cleanStart',
            default=False,
            help='do not restore previous IDE state (default: Off)')

        self.__options, self.__args = parser.parse_args()
        self.setupLogging()

        # The default exception handler can be replaced
        sys.excepthook = exceptionHook

        # Create global data singleton.
        # It's unlikely to throw any exceptions.
        globalData = GlobalData()
        globalData.version = VER

        # Loading settings - they have to be loaded before the application is
        # created. This is because the skin name is saved there.
        settings = Settings()
        populateSampleSkin()

        # Load the skin
        globalData.skin = Skin()
        globalData.skin.loadByName(settings['skin'])

        self.__delayedWarnings += settings.validateZoom(
            globalData.skin.minTextZoom, globalData.skin.minCFlowZoom)

        # QT on UBUNTU has a bug - the main menu bar does not generate the
        # 'aboutToHide' signal (though 'aboutToShow' is generated properly. This
        # prevents codimension working properly so the hack below disables the
        # global menu bar for codimension and makes it working properly.
        os.environ['QT_X11_NO_NATIVE_MENUBAR'] = '1'

        # Create QT application
        codimensionApp = CodimensionApplication(sys.argv, settings['style'])
        globalData.application = codimensionApp

        logging.debug('Starting codimension v.%s', VER)

        try:
            # Process command line arguments
            self.__projectFile = self.processCommandLineArgs()
        except Exception as exc:
            logging.error(str(exc))
            parser.print_help()
            return 1

        # Show splash screen
        self.__splash = SplashScreen()

        screenSize = codimensionApp.desktop().screenGeometry()
        globalData.screenWidth = screenSize.width()
        globalData.screenHeight = screenSize.height()

        self.__splash.showMessage('Importing packages...')
        from ui.mainwindow import CodimensionMainWindow

        self.__splash.showMessage('Generating main window...')
        mainWindow = CodimensionMainWindow(self.__splash, settings)
        codimensionApp.setMainWindow(mainWindow)
        globalData.mainWindow = mainWindow
        codimensionApp.lastWindowClosed.connect(codimensionApp.quit)

        mainWindow.show()
        mainWindow.restoreWindowPosition()
        mainWindow.restoreSplitterSizes()

        # Launch the user interface
        QTimer.singleShot(1, self.launchUserInterface)

        # Run the application main cycle
        retVal = codimensionApp.exec_()
        return retVal

    def launchUserInterface(self):
        """UI launchpad"""
        globalData = GlobalData()

        self.__splash.showMessage('Loading plugins...')
        globalData.pluginManager.load()

        settings = Settings()
        mainWindow = globalData.mainWindow
        mainWindow.getToolbar().setVisible(settings['showMainToolBar'])

        if not settings['showFSViewer']:
            mainWindow.projectViewer.onShowHide(True)

        needSignal = True
        if self.__options.cleanStart:
            # Codimension will not load anything.
            pass
        elif self.__projectFile:
            self.__splash.showMessage('Loading project...')
            globalData.project.loadProject(self.__projectFile)
            needSignal = False
        elif self.__args:
            # There are arguments and they are python files
            # The project should not be loaded but the files should
            # be opened
            for fName in self.__args:
                mainWindow.openFile(os.path.abspath(fName), -1)
        elif settings['projectLoaded']:
            if not settings['recentProjects']:
                # Some project was loaded but now it is not available.
                pass
            else:
                self.__splash.showMessage('Loading project...')
                if os.path.exists(settings['recentProjects'][0]):
                    globalData.project.loadProject(
                        settings['recentProjects'][0])
                    needSignal = False
                else:
                    self.__delayedWarnings.append(
                        'Cannot open the most recent project: ' +
                        settings['recentProjects'][0] +
                        '. Ignore and continue.')
        else:
            mainWindow.em.restoreTabs(settings.tabStatus)

        # Signal for triggering browsers layout
        if needSignal:
            globalData.project.sigProjectChanged.emit(
                CodimensionProject.CompleteProject)

        # The editors positions can be restored properly only when the editors
        # have actually been drawn. Otherwise the first visible line is
        # unknown. So, I load the project first and let object browsers
        # initialize themselves and then manually call the main window handler
        # to restore the editors. The last step is to connect the signal.
        mainWindow.onProjectChanged(CodimensionProject.CompleteProject)
        globalData.project.sigProjectChanged.connect(
            mainWindow.onProjectChanged)

        self.__splash.finish(globalData.mainWindow)
        self.__splash = None
        del self.__splash

        for message in self.__delayedWarnings:
            logging.warning(message)

        # Some startup time objects could be collected here. In my test runs
        # there were around 700 objects.
        gc.collect()


def exceptionHook(excType, excValue, tracebackObj):
    """Catches unhandled exceptions"""
    globalData = GlobalData()

    # Keyboard interrupt is a special case
    if issubclass(excType, KeyboardInterrupt):
        if globalData.application is not None:
            globalData.application.quit()
        return

    error = '%s: %s' % (excType.__name__, excValue)
    stackTraceString = ''.join(traceback.format_exception(excType, excValue,
                                                          tracebackObj))

    # Save the traceback to a file explicitly together with a log window
    # content.
    excptFileName = SETTINGS_DIR + 'unhandledexceptions.log'
    try:
        savedOK = True
        with open(excptFileName, 'a',
                  encoding=DEFAULT_ENCODING) as diskfile:
            diskfile.write('------ Unhandled exception report at ' +
                           str(datetime.datetime.now()) + '\n')
            diskfile.write('Traceback:\n')
            diskfile.write(stackTraceString)

            diskfile.write('Log window:\n')
            if globalData.mainWindow is not None:
                # i.e. the log window is available, save its content too
                logWindowContent = globalData.mainWindow.getLogViewerContent()
                logWindowContent = logWindowContent.strip()
                if logWindowContent:
                    diskfile.write(logWindowContent)
                    diskfile.write('\n')
                else:
                    diskfile.write('Nothing is there\n')
            else:
                diskfile.write('Has not been created yet\n')
            diskfile.write('------\n\n')
    except:
        savedOK = False

    # This output will be to a console if the application has not started yet
    # or to a log window otherwise.
    logging.error('Unhandled exception is caught\n%s', stackTraceString)

    # Display the message as a QT modal dialog box if the application
    # has started
    if globalData.application is not None:
        message = "<html><body>"
        if savedOK:
            message += "Stack trace and log window content saved in " + \
                       excptFileName + ".<br>"
        else:
            message += "Failed to save stack trace and log window " \
                       "content in " + excptFileName + ".<br>"

        lines = stackTraceString.split('\n')
        if len(lines) > 32:
            message += "First 32 lines of the stack trace " \
                       "(the rest is truncated):" \
                       "<pre>" + "\n".join(lines[:32]) + "<pre>"
        else:
            message += "Stack trace:" + \
                       "<pre>" + stackTraceString + "</pre>"
        message += "</body></html>"
        QMessageBox.critical(None, "Unhandled exception: " + error, message)
        globalData.application.exit(1)


def main():
    """The main entry point"""
    launcher = CodimensionUILauncher()
    retCode = launcher.main()

    # restore root logging handlers
    if launcher.rootLoggingHandlers:
        logging.root.handlers = launcher.rootLoggingHandlers

    logging.debug('Exiting codimension')
    sys.exit(retCode)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2015-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Various comment items on a virtual canvas"""

# pylint: disable=C0305
# pylint: disable=R0913

from html import escape
from ui.qt import (Qt, QPen, QBrush, QPainterPath, QGraphicsItem, QFont,
                   QGraphicsRectItem, QCursor, QDesktopServices, QUrl)
from utils.globals import GlobalData
from utils.misc import resolveLinkPath
from .auxitems import Connector
from .cellelement import CellElement
from .commentitems import CommentCellBase
from .colormixin import ColorMixin
from .iconmixin import IconMixin


class DocCellBase(CommentCellBase, ColorMixin, IconMixin, QGraphicsRectItem):

    """Base class for all doc cells"""

    def __init__(self, itemRef, cmlRef, canvas, x, y):
        CommentCellBase.__init__(self, itemRef, canvas, x, y)
        ColorMixin.__init__(self, None, canvas.settings.docLinkBGColor,
                            canvas.settings.docLinkFGColor,
                            canvas.settings.docLinkBorderColor,
                            colorSpec=cmlRef)
        QGraphicsRectItem.__init__(self, canvas.scopeRectangle)
        self.cmlRef = cmlRef

        # They all have the same icon
        if cmlRef.link is not None and cmlRef.anchor is not None:
            IconMixin.__init__(self, canvas, 'docanchor.svg',
                               'Jump to the documentation')
            self.iconItem.setCursor(QCursor(Qt.PointingHandCursor))
        else:
            if cmlRef.link is not None:
                IconMixin.__init__(self, canvas, 'doclink.svg',
                                   'Jump to the documentation')
                self.iconItem.setCursor(QCursor(Qt.PointingHandCursor))
            else:
                IconMixin.__init__(self, canvas, 'anchor.svg',
                                   'Documentation anchor')


        # They all are double clickable
        # This makes double click delivered
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)

    def mouseClickLinkIcon(self):
        """Follows the link"""
        if self.cmlRef.link is None:
            return

        # http://... an external browser will be invoked
        # https://... an external browser will be invoked
        # [file:]absolute path
        # [file:]relative path. The relative is tried to the current file
        #                       and then to the project root
        if self.cmlRef.link.startswith('http://') or \
           self.cmlRef.link.startswith('https://'):
            QDesktopServices.openUrl(QUrl(self.cmlRef.link))
            return

        fileName, anchorOrLine = resolveLinkPath(
            self.cmlRef.link, self.getEditor().getFileName())
        if fileName:
            GlobalData().mainWindow.openFile(fileName, anchorOrLine)

    def render(self):
        """Renders the cell"""
        settings = self.canvas.settings
        title = self.cmlRef.getTitle()
        self.setupText(self, customText=title, customReplacement='')

        if self.text:
            self.minWidth = self.textRect.width() + settings.hDocLinkPadding
            self.minHeight = self.textRect.height()
        else:
            self.minWidth = 0
            self.minHeight = self.iconItem.iconHeight()
            if settings.hidecomments:
                if title:
                    self.iconItem.setToolTip(self.iconItem.toolTip() +
                                             '<hr/><pre>' + escape(title) +
                                             '</pre>')

        self.minHeight += 2 * (settings.vCellPadding + \
                          settings.vDocLinkPadding)
        self.minWidth += 2 * (settings.hCellPadding + \
                         settings.hDocLinkPadding) + \
                         self.iconItem.iconWidth()

        self.height = self.minHeight
        self.width = self.minWidth
        return (self.width, self.height)

    def adjustWidth(self):
        """Used during rendering to adjust the width of the cell.

        The comment now can take some space on the left and the left hand
        side cell has to be rendered already.
        The width of this cell will take whatever is needed considering
        the comment shift to the left.
        """
        if self.kind == CellElement.ABOVE_DOC:
            return

        cellToTheLeft = self.canvas.cells[self.addr[1]][self.addr[0] - 1]
        if cellToTheLeft.kind != CellElement.CONNECTOR:
            # Not implemented yet
            return

        spareWidth = cellToTheLeft.width - cellToTheLeft.minWidth
        boxWidth = self.minWidth
        if spareWidth >= boxWidth:
            self.minWidth = 0
        else:
            self.minWidth = boxWidth - spareWidth
        self.width = self.minWidth

    def draw(self, scene, baseX, baseY):
        """Draws the cell"""
        self.baseX = baseX
        self.baseY = baseY

        self._setupConnector()
        scene.addItem(self.connector)

        settings = self.canvas.settings
        penWidth = settings.selectPenWidth - 1
        self.setRect(baseX + settings.hCellPadding - penWidth,
                     baseY + settings.vCellPadding - penWidth,
                     self.minWidth - 2 * settings.hCellPadding + 2 * penWidth,
                     self.minHeight - 2 * settings.vCellPadding + 2 * penWidth)
        scene.addItem(self)

        self.iconItem.setPos(
            baseX + settings.hCellPadding + settings.hDocLinkPadding,
            baseY + self.minHeight / 2 - self.iconItem.iconHeight() / 2)
        scene.addItem(self.iconItem)

    def paint(self, painter, option, widget):
        """Draws the independent comment"""
        del option      # unused argument
        del widget      # unused argument

        settings = self.canvas.settings
        painter.setPen(self.getPainterPen(self.isSelected(), self.borderColor))
        painter.setBrush(QBrush(self.bgColor))

        rectWidth = self.minWidth - 2 * settings.hCellPadding
        rectHeight = self.minHeight - 2 * settings.vCellPadding

        painter.drawRoundedRect(self.baseX + settings.hCellPadding,
                                self.baseY + settings.vCellPadding,
                                rectWidth, rectHeight, 0, 0)

        if self.text:
            # Draw the text in the rectangle
            font = QFont(settings.monoFont)
            font.setItalic(True)
            painter.setFont(font)
            pen = QPen(self.fgColor)
            painter.setPen(pen)
            painter.drawText(
                self._leftEdge + settings.hCellPadding +
                settings.hDocLinkPadding + self.iconItem.iconWidth() +
                settings.hDocLinkPadding,
                self.baseY + settings.vCellPadding + settings.vDocLinkPadding,
                self.textRect.width(), self.textRect.height(),
                Qt.AlignLeft, self.text)

    def getAbsPosRange(self):
        """Provides the absolute position range"""
        return self.cmlRef.getAbsPosRange()

    def getLineRange(self):
        """Provides the line range"""
        return self.cmlRef.getLineRange()

    def copyToClipboard(self):
        """Copies the item to a clipboard"""
        self._copyToClipboard(self.cmlRef.ref.parts)

    def getSelectTooltip(self):
        """Provides the tooltip"""
        return 'CML doc comment at ' + \
               CellElement.getLinesSuffix(self.getLineRange())


class IndependentDocCell(DocCellBase):

    """Represents a single independent CML doc comment"""

    def __init__(self, ref, canvas, x, y):
        DocCellBase.__init__(self, ref, ref, canvas, x, y)
        self.kind = CellElement.INDEPENDENT_DOC
        self.leadingForElse = False
        self.sideForElse = False

    def _setupConnector(self):
        """Sets the path for painting"""
        settings = self.canvas.settings

        cellToTheLeft = self.canvas.cells[self.addr[1]][self.addr[0] - 1]
        self._leftEdge = \
            cellToTheLeft.baseX + settings.mainLine + settings.hCellPadding

        # May be later the connector will look different for two cases below
        if self.leadingForElse:
            self.connector = Connector(
                self.canvas, self._leftEdge + settings.hCellPadding,
                self.baseY + self.minHeight / 2,
                cellToTheLeft.baseX + settings.mainLine,
                self.baseY + self.minHeight / 2)
        else:
            self.connector = Connector(
                self.canvas, self._leftEdge + settings.hCellPadding,
                self.baseY + self.minHeight / 2,
                cellToTheLeft.baseX + settings.mainLine,
                self.baseY + self.minHeight / 2)
        self.connector.penColor = self.borderColor
        self.connector.penWidth = settings.boxLineWidth



class LeadingDocCell(DocCellBase):

    """Represents a single leading CML doc comment"""

    def __init__(self, itemRef, cmlRef, canvas, x, y):
        DocCellBase.__init__(self, itemRef, cmlRef, canvas, x, y)
        self.kind = CellElement.LEADING_DOC

    def _setupConnector(self):
        """Sets the path for painting"""
        settings = self.canvas.settings

        # Bottom adjustment
        yShift = self.height - self.minHeight
        baseY = self.baseY + yShift

        cellToTheLeft = self.canvas.cells[self.addr[1]][self.addr[0] - 1]
        if cellToTheLeft.kind != CellElement.CONNECTOR:
            # not implemented yet
            self._leftEdge = self.baseX
        else:
            self._leftEdge = \
                cellToTheLeft.baseX + \
                settings.mainLine + settings.hCellPadding

        shift = self.hShift * 2 * settings.openGroupHSpacer
        self._leftEdge += shift

        self.connector = Connector(self.canvas, 0, 0, 0, 0)
        connectorPath = QPainterPath()
        connectorPath.moveTo(self._leftEdge + settings.hCellPadding,
                             baseY + self.minHeight / 2)
        connectorPath.lineTo(self._leftEdge,
                             baseY + self.minHeight / 2)
        connectorPath.lineTo(self._leftEdge - settings.hCellPadding,
                             baseY + self.minHeight + settings.vCellPadding)
        self.connector.setPath(connectorPath)
        self.connector.penColor = self.borderColor
        self.connector.penWidth = settings.boxLineWidth



class AboveDocCell(DocCellBase):

    """Represents a single leading doc link which is above certain blocks.

    Blocks are: try/except or for/else or while/else
    i.e. those which are scopes located in a single row
    """

    def __init__(self, itemRef, cmlRef, canvas, x, y):
        DocCellBase.__init__(self, itemRef, cmlRef, canvas, x, y)
        self.kind = CellElement.ABOVE_DOC
        self.needConnector = False

    def draw(self, scene, baseX, baseY):
        """Draws the cell"""
        if self.needConnector:
            settings = self.canvas.settings
            self.connector = Connector(
                self.canvas, baseX + settings.mainLine, baseY,
                baseX + settings.mainLine, baseY + self.height)
            scene.addItem(self.connector)

        DocCellBase.draw(self, scene,
                         baseX + self.canvas.settings.mainLine +
                         self.canvas.settings.hCellPadding, baseY)

    def _setupConnector(self):
        """Sets the path for painting"""
        settings = self.canvas.settings

        # Bottom adjustment
        yShift = self.height - self.minHeight
        baseY = self.baseY + yShift

        # Dirty hack: see the overriden draw() member: the baseX has already
        # been adjusted with mainLine and hCellPadding
        self._leftEdge = self.baseX

        self.connector = Connector(self.canvas, 0, 0, 0, 0)
        connectorPath = QPainterPath()
        connectorPath.moveTo(self._leftEdge + settings.hCellPadding,
                             baseY + self.minHeight / 2)
        connectorPath.lineTo(self._leftEdge,
                             baseY + self.minHeight / 2)
        connectorPath.lineTo(self._leftEdge - settings.hCellPadding,
                             baseY + self.minHeight + settings.vCellPadding)
        self.connector.setPath(connectorPath)
        self.connector.penColor = self.borderColor
        self.connector.penWidth = settings.boxLineWidth




# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""dot output (plain) parser"""

import os
import os.path
from utils.fileutils import getFileContent, saveToFile, makeTempFile
from utils.run import checkOutput


def splitWithQuotasRespect(line):
    """Splits the space separated values and respects quoted values"""
    def skipSpaces(line, startPos):
        """Provides index of first non-space char"""
        while startPos < len(line):
            if line[startPos] != ' ':
                return startPos
            startPos += 1
        return startPos

    def skipTillSpace(line, startPos):
        """Provides index of first space char"""
        while startPos < len(line):
            if line[startPos] == ' ':
                return startPos
            startPos += 1
        return startPos

    def skipTillClosedQuote(line, startPos):
        """Provides index of a closing quote"""
        while startPos < len(line):
            if line[startPos] == '"':
                if line[startPos - 1] == '\\':
                    startPos += 1
                    continue
                return startPos
            startPos += 1
        return startPos

    result = []

    index = 0
    while index < len(line):
        # skip spaces if so
        index = skipSpaces(line, index)
        if index >= len(line):
            return result

        # First symbol is not a space so it is a beginning of a value.
        startIndex = index
        if line[startIndex] == '"':
            # Beginning of a quoted value
            nextIndex = skipTillClosedQuote(line, startIndex + 1)
            result.append(line[startIndex + 1:nextIndex])
            index = nextIndex + 1
        else:
            nextIndex = skipTillSpace(line, startIndex)
            result.append(line[startIndex:nextIndex])
            index = nextIndex
    return result


class Graph():

    """Holds a description of a single graph"""

    def __init__(self):
        self.scale = 0.0
        self.width = 0.0
        self.height = 0.0

        self.vSpace = 10.0
        self.hSpace = 10.0

        self.edges = []
        self.nodes = []

    def normalize(self, scaleX, scaleY):
        """normalizes all the measures"""
        self.width = self.width * self.scale * scaleX
        self.height = self.height * self.scale * scaleY

        for edge in self.edges:
            edge.normalize(self, scaleX, scaleY)
        for node in self.nodes:
            node.normalize(self, scaleX, scaleY)

        # increase the size to have borders around
        self.width = self.width + 2.0 * self.hSpace
        self.height = self.height + 2.0 * self.vSpace

    def initFromLine(self, line):
        """Parses a line and initializes the values"""
        # graph scale width height
        # graph 1.000 57.306 10.833

        parts = line.strip().split()
        if len(parts) != 4:
            raise Exception("Unexpected number of parts in 'graph' statement")

        self.scale = float(parts[1].strip())
        self.width = float(parts[2].strip())
        self.height = float(parts[3].strip())


class Edge():

    """Holds a single graph edge description"""

    def __init__(self):
        self.tail  = ""
        self.head  = ""
        self.points = []

        self.label = ""
        self.labelX = 0.0
        self.labelY = 0.0

        self.style = ""
        self.color = ""

    def normalize(self, graph, scaleX, scaleY):
        """Scales to the screen coordinates"""
        self.labelX = self.labelX * graph.scale * scaleX + graph.hSpace
        self.labelY = graph.height - self.labelY * graph.scale * scaleY + \
                      graph.vSpace

        index = 0
        while index < len(self.points):
            # x
            self.points[index][0] = self.points[index][0] * graph.scale * scaleX
            self.points[index][0] = self.points[index][0] + graph.hSpace
            # y
            self.points[index][1] = self.points[index][1] * graph.scale * scaleY
            self.points[index][1] = graph.height - self.points[index][1] + \
                                    graph.vSpace
            index = index + 1

    def initFromLine(self, line):
        """Parses a line and initializes the values"""
        # edge tail head n x1 y1 .. xn yn [label xl yl] style color
        # edge "obj1" "obj2"
        #  4 29.806 10.389 29.208 10.236 28.375 10.000 27.722 9.833 solid black

        parts = splitWithQuotasRespect(line.strip())

        if len(parts) < 8:
            raise Exception("Unexpected number of parts in 'edge' "
                            "statement. Line: " + line)

        self.tail = parts[1]
        self.head = parts[2]

        numberOfPoints = int(parts[3])

        if len(parts) < (numberOfPoints * 2 + 5):
            raise Exception("Unexpected number of parts in 'edge' "
                            "statement. Line: " + line)

        point = 0
        while point < numberOfPoints:
            self.points.append([float(parts[point * 2 + 4]),
                                float(parts[point * 2 + 4 + 1])])
            point += 1

        # It is possible that there are spaces in the label
        self.label = ""
        self.labelX = 0.0
        self.labelY = 0.0
        self.style = ""
        self.color = ""

        # Strip the points description
        parts = parts[numberOfPoints * 2 + 4:]
        if len(parts) == 2:
            # There is no label
            self.style = parts[0]
            self.color = parts[1]
        else:
            self.label = parts[0]
            self.labelX = float(parts[1])
            self.labelY = float(parts[2])
            self.style = parts[3]
            self.color = parts[4]


class Node:

    """Holds a single node description"""

    def __init__(self):
        self.name = ""
        self.posX = 0.0
        self.posY = 0.0
        self.width = 0.0
        self.height = 0.0
        self.label = ""
        self.style = ""
        self.shape = ""
        self.color = ""
        self.fillcolor = ""

    def normalize(self, graph, scaleX, scaleY):
        """Scales to the screen coordinates"""
        self.posX = self.posX * graph.scale * scaleX + graph.hSpace
        self.posY = graph.height - self.posY * graph.scale * scaleY + \
                    graph.vSpace
        self.width = self.width * graph.scale * scaleX
        self.height = self.height * graph.scale * scaleY

    def initFromLine(self, line):
        """Parses a line and initializes the values"""
        # node name x y width height label style shape color fillcolor
        # node "/usr/local/vim-7.1/bin/vim"  30.542 10.583 2.388 0.500
        #      "/usr/local/vim-7.1/bin/vim" solid ellipse black lightgrey

        parts = splitWithQuotasRespect(line.strip())
        if len(parts) < 11:
            raise Exception("Unexpected number of parts in 'node' "
                            "statement. Line: " + line)

        self.name = parts[1]
        self.posX = float(parts[2].strip())
        self.posY = float(parts[3].strip())
        self.width = float(parts[4].strip())
        self.height = float(parts[5].strip())
        self.label = parts[6]
        self.style = parts[7].strip()
        self.shape = parts[8].strip()
        self.color = parts[9].strip()
        self.fillcolor = parts[10].strip()


def getGraphFromPlainDotData(content):
    """Parses data and builds a normalized graph"""
    graph = Graph()
    expectContinue = False
    combinedLine = ""
    for line in content.split('\n'):
        line = line.strip()
        if line == "":
            continue
        if line.endswith('\\'):
            combinedLine += line[:-1]
            expectContinue = True
            continue
        if expectContinue:
            expectContinue = False
        combinedLine += line

        if combinedLine.startswith("graph"):
            graph.initFromLine(combinedLine)
            combinedLine = ""
            continue
        if combinedLine.startswith("node"):
            node = Node()
            node.initFromLine(combinedLine)
            graph.nodes.append(node)
            combinedLine = ""
            continue
        if combinedLine.startswith("edge"):
            edge = Edge()
            edge.initFromLine(combinedLine)
            graph.edges.append(edge)
            combinedLine = ""
            continue
        if combinedLine.startswith("stop"):
            break

        raise Exception("Unexpected plain dot line: " + combinedLine)
    return graph


def getGraphFromPlainDotFile(fName):
    """Parses file and builds a normalized graph"""
    if not os.path.exists(fName):
        raise Exception("Cannot open " + fName)

    # Not optimal however I don't care at the moment.
    # In most of the cases this will be an interactive call.
    return getGraphFromPlainDotData(getFileContent(fName))


def getGraphFromDescrptionFile(fName):
    """Runs dot and then parses and builds normalized graph"""
    return getGraphFromPlainDotData(checkOutput(["dot", "-Tplain", fName]))


def getGraphFromDescriptionData(content):
    """Runs dot and then parses and builds normalized graph"""
    tempFileName = makeTempFile()
    saveToFile(tempFileName, content)

    try:
        graph = getGraphFromDescrptionFile(tempFileName)
    finally:
        os.unlink(tempFileName)
    return graph



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Text viewer tab widget"""

import os.path
import logging
from .qt import (QTextBrowser, QHBoxLayout, QWidget, Qt, pyqtSignal,
                 QDesktopServices)
from .mainwindowtabwidgetbase import MainWindowTabWidgetBase
from utils.misc import resolveLinkPath
from utils.globals import GlobalData


class TextViewer(QTextBrowser):

    """Text viewer"""

    sigEscapePressed = pyqtSignal()

    def __init__(self, parent=None):
        QTextBrowser.__init__(self, parent)

        self._parentWidget = parent
        self.setOpenExternalLinks(True)
        self.setOpenLinks(False)

        self.__copyAvailable = False
        self.copyAvailable.connect(self.__onCopyAvailable)
        self.anchorClicked.connect(self._onAnchorClicked)

    def __onCopyAvailable(self, available):
        """Triggered when copying is available"""
        self.__copyAvailable = available

    def _resolveLink(self, link):
        """Resolves the link to a file and optional anchor/line number"""
        scheme = link.scheme().lower()
        if scheme in ['http', 'https']:
            QDesktopServices.openUrl(link)
            return None, None

        if scheme == '':
            fileName = link.path()
        elif scheme == 'file':
            if link.isValid():
                fileName = link.path()
            else:
                logging.error('Invalid link: ' + link.errorString())
                return None, None
        elif scheme == 'action':
            if link.isValid():
                # The action is stored in the host part
                action = link.host()
                # The actions are predefined. I did not find a generic way
                # to find what the key is bound to
                if action.lower() == 'embedded-help':
                    GlobalData().mainWindow._onEmbeddedHelp()
                elif action.lower() == 'f1':
                    GlobalData().mainWindow.em.onHelp()
                elif action.lower() == 'project-cocumentation':
                    GlobalData().mainWindow.projectDocClicked()
                else:
                    # must be a keyboard shortcut
                    logging.error("Unsupported action '" + link.host() + "'")
            return None, None
        else:
            logging.error("Unsupported url scheme '" + link.scheme() +
                          "'. Supported schemes are 'http', 'https', 'file' "
                          "and an empty scheme for files")
            return None, None

        if not fileName:
            logging.error('Could not get a file name. Check the link format. '
                          'Valid examples: file:./relative/fname or '
                          'file:relative/fname or file:/absolute/fname or '
                          'file:///absolute/fname')
            return None, None

        fileName, anchorOrLine = resolveLinkPath(fileName,
                                                 self._parentWidget.getFileName())
        if anchorOrLine is None:
            if link.hasFragment():
                return fileName, link.fragment()
        return fileName, anchorOrLine

    def _onAnchorClicked(self, link):
        """Handles a URL click"""
        fileName, anchorOrLine = self._resolveLink(link)
        if fileName:
            GlobalData().mainWindow.openFile(fileName, anchorOrLine)

    def isCopyAvailable(self):
        """True if text copying is available"""
        return self.__copyAvailable

    def keyPressEvent(self, event):
        """Handles the key press events"""
        if event.key() == Qt.Key_Escape:
            self.sigEscapePressed.emit()
            event.accept()
        else:
            QTextBrowser.keyPressEvent(self, event)


class TextTabWidget(QWidget, MainWindowTabWidgetBase):

    """The widget which displays a RO HTML page"""

    sigEscapePressed = pyqtSignal()

    def __init__(self, parent=None):
        QWidget.__init__(self, parent)
        MainWindowTabWidgetBase.__init__(self)

        layout = QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)

        self.__editor = TextViewer(self)
        self.__editor.sigEscapePressed.connect(self.__onEsc)
        layout.addWidget(self.__editor)

        self.__fileName = ""
        self.__shortName = ""
        self.__encoding = "n/a"

    def __onEsc(self):
        """Triggered when Esc is pressed"""
        self.sigEscapePressed.emit()

    def setHTML(self, content):
        """Sets the content from the given string"""
        self.__editor.setHtml(content)

    def getHTML(self):
        """Provides the currently shown HTML"""
        return self.__editor.toHtml()

    def loadFormFile(self, path):
        """Loads the content from the given file"""
        f = open(path, 'r')
        content = f.read()
        f.close()
        self.setHTML(content)
        self.__fileName = path
        self.__shortName = os.path.basename(path)

    def getViewer(self):
        """Provides the QTextBrowser"""
        return self.__editor

    def setFocus(self):
        """Overridden setFocus"""
        self.__editor.setFocus()

    def isModified(self):
        """Tells if the file is modifed"""
        return False

    def getRWMode(self):
        """Tells the read/write mode"""
        return "RO"

    def getType(self):
        """Tells the widget type"""
        return MainWindowTabWidgetBase.HTMLViewer

    def getLanguage(self):
        """Tells the content language"""
        return "n/a"

    def getFileName(self):
        """Tells what file name of the widget"""
        return self.__fileName

    def setFileName(self, path):
        """Sets the file name"""
        self.__fileName = path
        self.__shortName = os.path.basename(path)

    def getEncoding(self):
        """Tells the content encoding"""
        return self.__encoding

    def setEncoding(self, newEncoding):
        """Sets the encoding - used for Diff files"""
        self.__encoding = newEncoding

    def getShortName(self):
        """Tells the display name"""
        return self.__shortName

    def setShortName(self, name):
        """Sets the display name"""
        self.__shortName = name



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Goto line widget implementation"""

from utils.pixmapcache import getIcon
from .qt import (QHBoxLayout, QToolButton, QLabel, QSizePolicy, QComboBox,
                 QWidget, QIntValidator, Qt)
from .mainwindowtabwidgetbase import MainWindowTabWidgetBase


class GotoLineWidget(QWidget):

    """goto bar widget"""

    maxHistory = 32

    def __init__(self, editorsManager, parent=None):

        QWidget.__init__(self, parent)
        self.editorsManager = editorsManager

        self.__gotoHistory = []

        # Common graphics items
        closeButton = QToolButton(self)
        closeButton.setToolTip("Click to close the dialog (ESC)")
        closeButton.setIcon(getIcon("close.png"))
        closeButton.clicked.connect(self.hide)

        lineLabel = QLabel(self)
        lineLabel.setText("Goto line:")

        self.linenumberEdit = QComboBox(self)
        self.linenumberEdit.setEditable(True)
        self.linenumberEdit.setInsertPolicy(QComboBox.InsertAtTop)
        self.linenumberEdit.setCompleter(None)
        self.linenumberEdit.setDuplicatesEnabled(False)
        sizePolicy = QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.linenumberEdit.sizePolicy().hasHeightForWidth())
        self.linenumberEdit.setSizePolicy(sizePolicy)
        self.validator = QIntValidator(1, 100000, self)
        self.linenumberEdit.setValidator(self.validator)
        self.linenumberEdit.editTextChanged.connect(self.__onEditTextChanged)
        self.linenumberEdit.lineEdit().returnPressed.connect(self.__onEnter)

        self.goButton = QToolButton(self)
        self.goButton.setToolTip("Click to jump to the line (ENTER)")
        self.goButton.setIcon(getIcon("gotoline.png"))
        self.goButton.setFocusPolicy(Qt.NoFocus)
        self.goButton.setEnabled(False)
        self.goButton.clicked.connect(self.__onGo)

        spacer = QWidget()
        spacer.setFixedWidth(1)

        horizontalLayout = QHBoxLayout(self)
        horizontalLayout.setContentsMargins(0, 0, 0, 0)

        horizontalLayout.addWidget(closeButton)
        horizontalLayout.addWidget(lineLabel)
        horizontalLayout.addWidget(self.linenumberEdit)
        horizontalLayout.addWidget(self.goButton)
        horizontalLayout.addWidget(spacer)

    def keyPressEvent(self, event):
        """Handles the key press events"""
        if event.key() == Qt.Key_Escape:
            activeWindow = self.editorsManager.currentWidget()
            if activeWindow:
                activeWindow.setFocus()
            event.accept()
            self.hide()

    def __updateHistory(self, txt):
        """Updates the combo history"""
        while txt in self.__gotoHistory:
            self.__gotoHistory.remove(txt)
        self.__gotoHistory = [txt] + self.__gotoHistory
        self.__gotoHistory = self.__gotoHistory[:GotoLineWidget.maxHistory]

        self.linenumberEdit.clear()
        self.linenumberEdit.addItems(self.__gotoHistory)

    def show(self):
        """Overriden show method"""
        self.linenumberEdit.lineEdit().selectAll()
        QWidget.show(self)
        self.activateWindow()

    def setFocus(self):
        """Overridded setFocus"""
        self.linenumberEdit.setFocus()

    def updateStatus(self):
        """Triggered when the current tab is changed"""
        currentWidget = self.editorsManager.currentWidget()
        status = currentWidget.getType() in \
            [MainWindowTabWidgetBase.PlainTextEditor,
             MainWindowTabWidgetBase.VCSAnnotateViewer]
        self.linenumberEdit.setEnabled(status)
        self.goButton.setEnabled(status and
                                 self.linenumberEdit.currentText() != "")

    def __onGo(self):
        """Triggered when the 'Go!' button is clicked"""
        if self.linenumberEdit.currentText() == "":
            return

        currentWidget = self.editorsManager.currentWidget()
        if not currentWidget.getType() in \
           [MainWindowTabWidgetBase.PlainTextEditor,
            MainWindowTabWidgetBase.VCSAnnotateViewer]:
            return

        txt = self.linenumberEdit.currentText()
        self.__updateHistory(txt)
        editor = currentWidget.getEditor()
        line = min(int(txt), len(editor.lines)) - 1

        editor.cursorPosition = line, 0
        editor.ensureLineOnScreen(line)
        currentWidget.setFocus()

    def __onEditTextChanged(self, text):
        """Triggered when the text has been changed"""
        self.goButton.setEnabled(text != "")

    def __onEnter(self):
        """Triggered when 'Enter' or 'Return' is clicked"""
        self.__onGo()

    def selectAll(self):
        """Selects the line edit content"""
        self.linenumberEdit.lineEdit().selectAll()



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# The file was taken from eric 4 and adopted for codimension.
# Original copyright:
# Copyright (c) 2004 - 2014 Detlev Offenbach <detlev@die-offenbachs.de>
#


"""Module implementing utilities to get a password and/or the current user name

getpass(prompt) - prompt for a password, with echo turned off
getuser() - get the user name from the environment or password database

This module is a replacement for the one found in the Python distribution. It
is to provide a debugger compatible variant of the functions mentioned above.
"""

__all__ = ["getpass", "getuser"]


def getuser():
    """Function to get the username from the environment or password database.

    First try various environment variables, then the password
    database.  This works on Windows as long as USERNAME is set.
    """
    # this is copied from the oroginal getpass.py

    import os

    for name in ('LOGNAME', 'USER', 'LNAME', 'USERNAME'):
        user = os.environ.get(name)
        if user:
            return user

    # If this fails, the exception will "explain" why
    import pwd
    return pwd.getpwuid(os.getuid())[0]


def getpass(prompt='Password: '):
    """Function to prompt for a password, with echo turned off"""
    return input(prompt, False)


unix_getpass = getpass
win_getpass = getpass
default_getpass = getpass



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2015-2019  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Various minimized items for a virtual canvas"""

# pylint: disable=C0305

from sys import maxsize
from html import escape
from ui.qt import Qt, QBrush, QGraphicsRectItem, QGraphicsItem, QPainterPath
from .auxitems import Connector
from .iconmixin import IconMixin
from .cellelement import CellElement
from .routines import distance


class MinimizedCellBase(CellElement, IconMixin, QGraphicsRectItem):

    """Base for all minimized cells"""

    def __init__(self, iconFileName, ref, canvas, x, y):
        CellElement.__init__(self, ref, canvas, x, y)
        IconMixin.__init__(self, canvas, iconFileName)
        QGraphicsRectItem.__init__(self, canvas.scopeRectangle)

        # Visually the icon looks a bit too big so reduce the size to 80%
        self.iconItem.setIconHeight(self.iconItem.iconHeight() * 0.8)

        self.rectWidth = None
        self.rectHeight = None
        self.connector = None
        self.rectRadius = None

        # To make double click delivered
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)

    def renderCell(self, hIconPadding, vIconPadding):
        """Renders the cell"""
        settings = self.canvas.settings

        if self.kind == CellElement.EXCEPT_MINIMIZED:
            self.rectRadius = settings.hiddenExceptRectRadius
        else:
            self.rectRadius = settings.hiddenCommentRectRadius

        self.rectWidth = self.iconItem.iconWidth() + 2 * hIconPadding
        self.rectHeight = self.iconItem.iconHeight() + 2 * vIconPadding

        self.minWidth = self.rectWidth + 2 * settings.hCellPadding
        self.minHeight = self.rectHeight + 2 * settings.vCellPadding

        self.height = self.minHeight
        self.width = self.minWidth
        return (self.width, self.height)

    def drawCell(self, scene, baseX, baseY, hIconPadding,
                 setupConnector):
        """Draws the cell"""
        self.baseX = baseX
        self.baseY = baseY

        # derived class method; it uses self.baseX and self.baseY
        setupConnector()
        scene.addItem(self.connector)

        # xPos matches the connector (which could be drawn in any direction)
        xPos = max(self.connector.getFirstPoint()[0],
                   self.connector.getFirstPoint()[0])

        settings = self.canvas.settings
        penWidth = settings.selectPenWidth - 1
        self.setRect(xPos - penWidth,
                     baseY + settings.vCellPadding - penWidth,
                     self.rectWidth + 2 * penWidth,
                     self.rectHeight + 2 * penWidth)
        scene.addItem(self)

        self.iconItem.setPos(
            xPos + hIconPadding,
            baseY + self.minHeight / 2 - self.iconItem.iconHeight() / 2)
        scene.addItem(self.iconItem)

    def paintCell(self, painter, bgColor, borderColor,
                  option, widget):
        """Paints the comment"""
        del option
        del widget

        settings = self.canvas.settings
        painter.setPen(self.getPainterPen(self.isSelected(), borderColor))
        painter.setBrush(QBrush(bgColor))

        # xPos matches the connector (which could be drawn in any direction)
        xPos = max(self.connector.getFirstPoint()[0],
                   self.connector.getFirstPoint()[0])

        painter.drawRoundedRect(xPos,
                                self.baseY + settings.vCellPadding,
                                self.rectWidth, self.rectHeight,
                                self.rectRadius, self.rectRadius)


class MinimizedExceptCell(MinimizedCellBase):

    """Represents a minimized except block"""

    def __init__(self, ref, canvas, x, y):
        MinimizedCellBase.__init__(self, 'hiddenexcept.svg', ref, canvas, x, y)
        self.kind = CellElement.EXCEPT_MINIMIZED

        self.__setTooltip()

    def __setTooltip(self):
        """Sets the item tooltip"""
        parts = []
        for part in self.ref.exceptParts:
            lines = part.getDisplayValue().splitlines()
            if len(lines) > 1:
                parts.append('except: ' + lines[0] + '...')
            elif len(lines) == 1:
                parts.append('except: ' + lines[0])
            else:
                parts.append('except:')
        self.iconItem.setToolTip('<pre>' + escape('\n'.join(parts)) + '</pre>')

    def __setupConnector(self):
        """Sets the connector"""
        settings = self.canvas.settings

        cellToTheLeft = self.canvas.cells[self.addr[1]][self.addr[0] - 1]
        leftEdge = cellToTheLeft.baseX + cellToTheLeft.minWidth
        height = min(self.minHeight / 2, cellToTheLeft.minHeight / 2)

        self.connector = Connector(
            self.canvas, leftEdge + settings.hCellPadding,
            self.baseY + height,
            cellToTheLeft.baseX +
            cellToTheLeft.minWidth - settings.hCellPadding,
            self.baseY + height)

        self.connector.penStyle = Qt.DotLine

    def render(self):
        """Renders the cell"""
        return self.renderCell(self.canvas.settings.hHiddenExceptPadding,
                               self.canvas.settings.vHiddenExceptPadding)

    def draw(self, scene, baseX, baseY):
        """Draws the cell"""
        self.drawCell(scene, baseX, baseY,
                      self.canvas.settings.hHiddenExceptPadding,
                      self.__setupConnector)

    def paint(self, painter, option, widget):
        """Paints the cell"""
        self.paintCell(painter,
                       self.canvas.settings.hiddenExceptBGColor,
                       self.canvas.settings.hiddenExceptBorderColor,
                       option, widget)

    def mouseDoubleClickEvent(self, event):
        """Jump to the appropriate line in the text editor"""
        firstExcept = self.ref.exceptParts[0]
        CellElement.mouseDoubleClickEvent(self, event,
                                          line=firstExcept.body.beginLine,
                                          pos=firstExcept.body.beginPos)

    def getLineRange(self):
        """Provides the line range"""
        firstLineRange = self.ref.exceptParts[0].getLineRange()
        lastLineRange = self.ref.exceptParts[-1].getLineRange()
        return [firstLineRange[0], lastLineRange[1]]

    def getAbsPosRange(self):
        """Provides the absolute position range"""
        firstExcept = self.ref.exceptParts[0]
        lastExcept = self.ref.exceptParts[-1]
        return [firstExcept.begin, lastExcept.end]

    def getSelectTooltip(self):
        """Provides the tooltip"""
        lineRange = self.getLineRange()
        count = len(self.ref.exceptParts)
        if count == 1:
            return 'Minimized except block at ' + \
                   CellElement.getLinesSuffix(lineRange)
        return str(count) + ' minimized except blocks at ' + \
               CellElement.getLinesSuffix(lineRange)


class MinimizedCommentBase:

    """Base class for all the minimized comment cells"""

    def __init__(self):
        pass

    @staticmethod
    def selectTooltip(lineRange):
        """Provides the tooltip"""
        return 'Minimized comment at ' + CellElement.getLinesSuffix(lineRange)



class MinimizedIndependentCommentCell(MinimizedCommentBase, MinimizedCellBase):

    """Represents a minimized independent comment"""

    def __init__(self, ref, canvas, x, y):
        MinimizedCommentBase.__init__(self)
        MinimizedCellBase.__init__(self, 'hiddencomment.svg',
                                   ref, canvas, x, y)
        self.kind = CellElement.INDEPENDENT_MINIMIZED_COMMENT

        self.__setTooltip()

        self.leadingForElse = False
        self.sideForElse = False

    def __setTooltip(self):
        """Sets the item tooltip"""
        displayValue = self.ref.getDisplayValue()
        if displayValue:
            self.iconItem.setToolTip('<pre>' + escape(displayValue) + '</pre>')

    def __setupConnector(self):
        """Prepares the connector"""
        settings = self.canvas.settings

        cellToTheLeft = self.canvas.cells[self.addr[1]][self.addr[0] - 1]
        leftEdge = \
            cellToTheLeft.baseX + settings.mainLine + settings.hCellPadding

        if self.leadingForElse:
            self.connector = Connector(
                self.canvas, leftEdge + settings.hCellPadding,
                self.baseY + self.minHeight / 2,
                cellToTheLeft.baseX + settings.mainLine,
                self.baseY + self.minHeight / 2)
        else:
            self.connector = Connector(
                self.canvas, leftEdge + settings.hCellPadding,
                self.baseY + self.minHeight / 2,
                cellToTheLeft.baseX + settings.mainLine,
                self.baseY + self.minHeight / 2)
        self.connector.penColor = settings.hiddenCommentBorderColor
        self.connector.penWidth = settings.boxLineWidth

    def render(self):
        """Renders the cell"""
        return self.renderCell(self.canvas.settings.hHiddenCommentPadding,
                               self.canvas.settings.vHiddenCommentPadding)

    def draw(self, scene, baseX, baseY):
        """Draws the cell"""
        self.drawCell(scene, baseX, baseY,
                      self.canvas.settings.hHiddenCommentPadding,
                      self.__setupConnector)

    def paint(self, painter, option, widget):
        """Draws the independent comment"""
        self.paintCell(painter,
                       self.canvas.settings.hiddenCommentBGColor,
                       self.canvas.settings.hiddenCommentBorderColor,
                       option, widget)

    def adjustWidth(self):
        """No need to adjust the width"""

    def mouseDoubleClickEvent(self, event):
        """Jump to the appropriate line in the text editor"""
        # Needed custom because this item is used for ifs 'else' side comment
        CellElement.mouseDoubleClickEvent(self, event,
                                          pos=self.ref.beginPos)

    def getLineRange(self):
        """Provides the line range"""
        return self.ref.getLineRange()

    def getAbsPosRange(self):
        """Provides the absolute position range"""
        return [self.ref.begin, self.ref.end]

    def getSelectTooltip(self):
        """Provides the tooltip"""
        return MinimizedCommentBase.selectTooltip(self.getLineRange())


class MinimizedLeadingCommentCell(MinimizedCommentBase, MinimizedCellBase):

    """Represents a minimized leading comment"""

    def __init__(self, ref, canvas, x, y):
        MinimizedCommentBase.__init__(self)
        MinimizedCellBase.__init__(self, 'hiddencomment.svg',
                                   ref, canvas, x, y)
        self.kind = CellElement.LEADING_MINIMIZED_COMMENT

        self.__setTooltip()

    def __setTooltip(self):
        """Sets the item tooltip"""
        displayValue = self.ref.leadingComment.getDisplayValue()
        if displayValue:
            self.iconItem.setToolTip('<pre>' + escape(displayValue) + '</pre>')

    def __setupConnector(self):
        """Prepares the connector"""
        settings = self.canvas.settings

        cellToTheLeft = self.canvas.cells[self.addr[1]][self.addr[0] - 1]
        if cellToTheLeft.kind != CellElement.CONNECTOR:
            leftEdge = self.baseX
        else:
            leftEdge = \
                cellToTheLeft.baseX + settings.mainLine + settings.hCellPadding

        shift = self.hShift * 2 * settings.openGroupHSpacer
        leftEdge += shift

        self.connector = Connector(self.canvas, 0, 0, 0, 0)
        connectorPath = QPainterPath()
        connectorPath.moveTo(leftEdge + settings.hCellPadding,
                             self.baseY + self.minHeight / 2)
        connectorPath.lineTo(leftEdge, self.baseY + self.minHeight / 2)
        connectorPath.lineTo(leftEdge - settings.hCellPadding,
                             self.baseY + self.minHeight +
                             settings.vCellPadding)
        self.connector.setPath(connectorPath)
        self.connector.penColor = settings.hiddenCommentBorderColor
        self.connector.penWidth = settings.boxLineWidth

    def render(self):
        """Renders the cell"""
        return self.renderCell(self.canvas.settings.hHiddenCommentPadding,
                               self.canvas.settings.vHiddenCommentPadding)

    def draw(self, scene, baseX, baseY):
        """Draws the cell"""
        self.drawCell(scene, baseX, baseY,
                      self.canvas.settings.hHiddenCommentPadding,
                      self.__setupConnector)

    def paint(self, painter, option, widget):
        """Draws the independent comment"""
        self.paintCell(painter,
                       self.canvas.settings.hiddenCommentBGColor,
                       self.canvas.settings.hiddenCommentBorderColor,
                       option, widget)

    def adjustWidth(self):
        """Adjust the width"""
        cellToTheLeft = self.canvas.cells[self.addr[1]][self.addr[0] - 1]
        if cellToTheLeft.kind != CellElement.CONNECTOR:
            return

        settings = self.canvas.settings
        spareWidth = cellToTheLeft.width - cellToTheLeft.minWidth
        boxWidth = self.minWidth - 2 * settings.hCellPadding
        if spareWidth >= boxWidth:
            self.minWidth = 0
        else:
            self.minWidth = boxWidth - spareWidth
        self.width = self.minWidth

    def getLineRange(self):
        """Provides the line range"""
        return self.ref.leadingComment.getLineRange()

    def getAbsPosRange(self):
        """Provides the absolute position range"""
        return [self.ref.leadingComment.begin, self.ref.leadingComment.end]

    def getSelectTooltip(self):
        """Provides the tooltip"""
        return MinimizedCommentBase.selectTooltip(self.getLineRange())



class MinimizedAboveCommentCell(MinimizedCommentBase, MinimizedCellBase):

    """Represents a minimized above comment"""

    def __init__(self, ref, canvas, x, y):
        MinimizedCommentBase.__init__(self)
        MinimizedCellBase.__init__(self, 'hiddencomment.svg',
                                   ref, canvas, x, y)
        self.kind = CellElement.ABOVE_MINIMIZED_COMMENT
        self.needConnector = False
        self.vConnector = None
        self.__setTooltip()

    def __setTooltip(self):
        """Sets the item tooltip"""
        displayValue = self.ref.leadingComment.getDisplayValue()
        if displayValue:
            self.iconItem.setToolTip('<pre>' + escape(displayValue) + '</pre>')

    def __setupConnector(self):
        """Prepares the connector"""
        settings = self.canvas.settings

        leftEdge = \
            self.baseX + settings.mainLine + settings.hCellPadding

        self.connector = Connector(self.canvas, 0, 0, 0, 0)
        connectorPath = QPainterPath()
        connectorPath.moveTo(leftEdge + settings.hCellPadding,
                             self.baseY + self.minHeight / 2)
        connectorPath.lineTo(leftEdge,
                             self.baseY + self.minHeight / 2)
        connectorPath.lineTo(leftEdge - settings.hCellPadding,
                             self.baseY + self.minHeight +
                             settings.vCellPadding)
        self.connector.setPath(connectorPath)
        self.connector.penColor = settings.commentBorderColor
        self.connector.penWidth = settings.boxLineWidth

    def render(self):
        """Renders the cell"""
        self.renderCell(self.canvas.settings.hHiddenCommentPadding,
                        self.canvas.settings.vHiddenCommentPadding)
        self.minWidth += self.canvas.settings.mainLine
        self.minWidth += self.canvas.settings.hCellPadding
        self.width = self.minWidth
        return (self.width, self.height)

    def draw(self, scene, baseX, baseY):
        """Draws the cell"""
        settings = self.canvas.settings
        if self.needConnector:
            self.vConnector = Connector(
                self.canvas, baseX + settings.mainLine, baseY,
                baseX + settings.mainLine, baseY + self.height)
            scene.addItem(self.vConnector)

        self.drawCell(scene, baseX, baseY,
                      settings.hHiddenCommentPadding,
                      self.__setupConnector)

    def paint(self, painter, option, widget):
        """Draws the independent comment"""
        settings = self.canvas.settings
        self.paintCell(painter,
                       settings.hiddenCommentBGColor,
                       settings.hiddenCommentBorderColor,
                       option, widget)

    def adjustWidth(self):
        """No need to adjust the width"""

    def getLineRange(self):
        """Provides the line range"""
        return self.ref.leadingComment.getLineRange()

    def getAbsPosRange(self):
        """Provides the absolute position range"""
        return [self.ref.leadingComment.begin, self.ref.leadingComment.end]

    def getSelectTooltip(self):
        """Provides the tooltip"""
        return MinimizedCommentBase.selectTooltip(self.getLineRange())



class MinimizedSideCommentCell(MinimizedCommentBase, MinimizedCellBase):

    """Represents a minimized side comment"""

    def __init__(self, ref, canvas, x, y):
        MinimizedCommentBase.__init__(self)
        MinimizedCellBase.__init__(self, 'hiddencomment.svg',
                                   ref, canvas, x, y)
        self.kind = CellElement.SIDE_MINIMIZED_COMMENT

        self.__setTooltip()

    def __setTooltip(self):
        """Sets the item tooltip"""
        displayValue = self.ref.sideComment.getDisplayValue()
        if displayValue:
            self.iconItem.setToolTip('<pre>' + escape(displayValue) + '</pre>')

    def __setupConnector(self):
        """Prepares the connector"""
        settings = self.canvas.settings

        cellToTheLeft = self.canvas.cells[self.addr[1]][self.addr[0] - 1]

        if cellToTheLeft.kind == CellElement.CONNECTOR:
            ifCell = None
            for cell in self.canvas.cells[self.addr[1]]:
                if cell.kind == CellElement.IF:
                    ifCell = cell
                    break
        elif cellToTheLeft.kind == CellElement.IF:
            ifCell = cellToTheLeft
        else:
            ifCell = None

        if cellToTheLeft.kind == CellElement.CONNECTOR:
            # 'if' or 'elif' side comment when there is a connector
            leftEdge = \
                cellToTheLeft.baseX + settings.mainLine + settings.hCellPadding

            index = self.addr[0] - 1
            while self.canvas.cells[self.addr[1]][index].kind == \
                    CellElement.CONNECTOR:
                index -= 1

            yPos = self.baseY + settings.vCellPadding + \
                   settings.ifSideCommentVShift

            self.connector = Connector(
                self.canvas, leftEdge + settings.hCellPadding,
                yPos,
                ifCell.baseX + ifCell.minWidth - settings.hCellPadding,
                yPos)
        else:
            # Regular box or 'if' without an rhs connector
            if ifCell:
                leftEdge = ifCell.baseX + ifCell.minWidth
            else:
                leftEdge = cellToTheLeft.baseX + cellToTheLeft.minWidth

            if ifCell:
                yPos = self.baseY + settings.vCellPadding + \
                       settings.ifSideCommentVShift
                self.connector = Connector(
                    self.canvas, leftEdge + settings.hCellPadding,
                    yPos,
                    ifCell.baseX + ifCell.minWidth - settings.hCellPadding,
                    yPos)
            else:
                height = min(self.minHeight / 2, cellToTheLeft.minHeight / 2)
                self.connector = Connector(
                    self.canvas, leftEdge + settings.hCellPadding,
                    self.baseY + height,
                    cellToTheLeft.baseX +
                    cellToTheLeft.minWidth - settings.hCellPadding,
                    self.baseY + height)

        self.connector.penColor = settings.commentBorderColor
        self.connector.penWidth = settings.boxLineWidth

    def render(self):
        """Renders the cell"""
        return self.renderCell(self.canvas.settings.hHiddenCommentPadding,
                               self.canvas.settings.vHiddenCommentPadding)

    def draw(self, scene, baseX, baseY):
        """Draws the cell"""
        self.drawCell(scene, baseX, baseY,
                      self.canvas.settings.hHiddenCommentPadding,
                      self.__setupConnector)

    def paint(self, painter, option, widget):
        """Draws the independent comment"""
        self.paintCell(painter,
                       self.canvas.settings.hiddenCommentBGColor,
                       self.canvas.settings.hiddenCommentBorderColor,
                       option, widget)

    def adjustWidth(self):
        """Adjusting the cell width"""
        settings = self.canvas.settings
        cellToTheLeft = self.canvas.cells[self.addr[1]][self.addr[0] - 1]
        spareWidth = cellToTheLeft.width - cellToTheLeft.minWidth
        if spareWidth > 0:
            self.minWidth = max(0, self.minWidth - spareWidth)
        self.width = self.minWidth

    def mouseDoubleClickEvent(self, event):
        """Jump to the appropriate line in the text editor"""
        CellElement.mouseDoubleClickEvent(self, event,
                                          pos=self.ref.sideComment.beginPos)

    def getLineRange(self):
        """Provides the line range"""
        return self.ref.sideComment.getLineRange()

    def getAbsPosRange(self):
        """Provides the absolute position range"""
        return self.ref.sideComment.getAbsPosRange()

    def getSelectTooltip(self):
        """Provides the tooltip"""
        return MinimizedCommentBase.selectTooltip(self.getLineRange())

    def getDistance(self, absPos):
        """Provides a distance between the absPos and the item"""
        retval = maxsize
        for part in self.ref.sideComment.parts:
            # +1 is for finishing \n character
            dist = distance(absPos, part.begin, part.end + 1)
            if dist == 0:
                return 0
            retval = min(retval, dist)
        return retval

    def getLineDistance(self, line):
        """Provides a distance between the line and the item"""
        retval = maxsize
        for part in self.ref.sideComment.parts:
            dist = distance(line, part.beginLine, part.endLine)
            if dist == 0:
                return 0
            retval = min(retval, dist)
        return retval



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# The file was taken from eric 4.4.3 and adopted for codimension.
# Original copyright:
# Copyright (c) 2007 - 2010 Detlev Offenbach <detlev@die-offenbachs.de>
#

"""File system browser with module browsing capabilities"""

from utils.pixmapcache import getIcon
from utils.globals import GlobalData
from .filesystembrowsermodel import FileSystemBrowserModel
from .filesbrowserbase import FilesBrowser


class FileSystemBrowser(FilesBrowser):

    """File system tree browser"""

    def __init__(self, parent=None):
        FilesBrowser.__init__(self, FileSystemBrowserModel(), False, parent)

        self.setWindowTitle('Filesystem browser')
        self.setWindowIcon(getIcon('icon.png'))

        GlobalData().project.sigFSChanged.connect(self._onFSChanged)

    def removeToplevelDir(self):
        """Handles the Remove from toplevel popup menu entry"""
        index = self.currentIndex()
        dname = self.model().item(index).getPath()
        GlobalData().project.removeTopLevelDir(dname)
        sindex = self.model().mapToSource(index)
        self.model().sourceModel().removeTopLevelDir(sindex)

    def addToplevelDir(self):
        """Handles the Add as toplevel directory popup menu entry"""
        index = self.currentIndex()
        dname = self.model().item(index).getPath()
        GlobalData().project.addTopLevelDir(dname)
        self.model().sourceModel().addTopLevelDir(dname)
        self.layoutDisplay()

    def reload(self):
        """Reloads the filesystem view"""
        self.model().sourceModel().populateModel()
        self.model().beginResetModel()
        self.model().endResetModel()
        self.layoutDisplay()



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2017 Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""Variable filters"""

# A set of functions to filter variables in the debugger.
# Each function should return True if the variable matches the filter.
# The filtration is done on the IDE side.
# Both name and type are strings


def filterLocalVariables(isGlobal, varName, varType):
    """Filters out the local variables"""
    del varName     # unused argument
    del varType     # unused argument
    return not isGlobal


def filterGlobalVariables(isGlobal, varName, varType):
    """Filters out the global variables"""
    del varName     # unused argument
    del varType     # unused argument
    return isGlobal


def filterHiddenAttributes(isGlobal, varName, varType):
    """Filters out the hidden attributes"""
    del isGlobal    # unused argument
    del varType     # unused argument
    return varName.startswith('__') and varName.endswith('__')


def filterTypes(isGlobal, varName, varType):
    """Filters out types"""
    del isGlobal    # unused argument
    del varName     # unused argument
    return varType.lower() == 'type'


def filterClassMethods(isGlobal, varName, varType):
    """Filters out methods"""
    del isGlobal    # unused argument
    del varName     # unused argument
    return varType.lower() == 'class method'


def filterFunctions(isGlobal, varName, varType):
    """Filters out functions"""
    del isGlobal    # unused argument
    del varName     # unused argument
    return varType.lower() == 'function'


def filterBuiltinFunctions(isGlobal, varName, varType):
    """Filters out builtin functions"""
    del isGlobal    # unused argument
    del varName     # unused argument
    return varType.lower() in ['method-wrapper', 'builtin function']


def filterModules(isGlobal, varName, varType):
    """Filters out modules"""
    del isGlobal    # unused argument
    del varName     # unused argument
    return varType.lower() == 'module'


def filterNones(isGlobal, varName, varType):
    """Filters out None-type variables"""
    del isGlobal    # unused argument
    del varName     # unused argument
    return varType.lower() == 'none'



VARIABLE_FILTERS = [
    ['Local Variables', 'dbgfltlocal', filterLocalVariables],
    ['Global Variables', 'dbgfltglobal', filterGlobalVariables],
    ['Hidden Attributes', 'dbgflthidden', filterHiddenAttributes],
    ['Types', 'dbgflttype', filterTypes],
    ['Class Methods', 'dbgfltmethod', filterClassMethods],
    ['Functions', 'dbgfltfunc', filterFunctions],
    ['Builtin Functions', 'dbgfltbuiltin', filterBuiltinFunctions],
    ['Modules', 'dbgfltmodule', filterModules],
    ['No Types', 'dbgfltnotype', filterNones]]



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2017  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# The file was taken from eric 4 and adopted for codimension.
# Original copyright:
# Copyright (c) 2002 - 2012 Detlev Offenbach <detlev@die-offenbachs.de>
#

"""Module defining the debug protocol tokens"""

# Distinctive return codes
KILLED = -100000
DISCONNECTED = -200000
FAILED_TO_START = -300000
SYNTAX_ERROR_AT_START = -400000
STOPPED_BY_REQUEST = -500000
UNHANDLED_EXCEPTION = -600000

METHOD_STDOUT = 'Stdout'
METHOD_STDERR = 'Stderr'
METHOD_STDIN = 'Stdin'

METHOD_CLIENT_OUTPUT = 'ClientOutput'
METHOD_THREAD_LIST = 'ThreadList'                       # to/from IDE
METHOD_THREAD_SET = 'ThreadSet'                         # to/from IDE
METHOD_VARIABLES = 'Variables'                          # to/from IDE
METHOD_VARIABLE = 'Variable'                            # to/from IDE
METHOD_FORK_TO = 'ForkTo'                               # to/from IDE
METHOD_CALL_TRACE = 'CallTrace'                         # to/from IDE

METHOD_DEBUG_STARTUP = 'DebugStartup'                   # to IDE
METHOD_LINE = 'Line'                                    # to IDE
METHOD_EXCEPTION = 'Exception'                          # to IDE
METHOD_STACK = 'Stack'                                  # to IDE
METHOD_BP_CONDITION_ERROR = 'BPConditionError'          # to IDE
METHOD_WP_CONDITION_ERROR = 'WPConditionError'          # to IDE
METHOD_CLEAR_BP = 'ClearBreakpoint'                     # to IDE
METHOD_CLEAR_WP = 'ClearWatchpoint'                     # to IDE
METHOD_SYNTAX_ERROR = 'SyntaxError'                     # to IDE
METHOD_SIGNAL = 'Signal'                                # to IDE
METHOD_EXEC_STATEMENT_ERROR = 'ExecStatementError'      # to IDE
METHOD_EXEC_STATEMENT_OUTPUT = 'ExecStatementOutput'    # to IDE

METHOD_CONTINUE = 'Continue'                            # from IDE
METHOD_STEP_QUIT = 'StepQuit'                           # from IDE
METHOD_STEP_OUT = 'StepOut'                             # from IDE
METHOD_STEP_OVER = 'StepOver'                           # from IDE
METHOD_STEP = 'Step'                                    # from IDE
METHOD_MOVE_IP = 'MoveIP'                               # from IDE
METHOD_SET_BP = 'SetBreakpoint'                         # from IDE
METHOD_BP_ENABLE = 'EnableBreakpoint'                   # from IDE
METHOD_BP_IGNORE = 'IgnoreBreakpoint'                   # from IDE
METHOD_SET_WP = 'SetWatchpoint'                         # from IDE
METHOD_WP_ENABLE = 'EnableWatchpoint'                   # from IDE
METHOD_WP_IGNORE = 'IgnoreWatchpoint'                   # from IDE
METHOD_SET_ENVIRONMENT = 'SetEnvironment'               # from IDE
METHOD_EXECUTE_STATEMENT = 'ExecuteStatement'           # from IDE
METHOD_SHUTDOWN = 'Shutdown'                            # from IDE
METHOD_SET_FILTER = 'SetFilter'                         # from IDE

METHOD_PROC_ID_INFO = 'ProcIDInfo'
METHOD_PROLOGUE_CONTINUE = 'PrologueContinue'
METHOD_EPILOGUE_EXIT = 'EpilogueExit'
METHOD_EPILOGUE_EXIT_CODE = 'EpilogueExitCode'

VAR_TYPE_DISP_STRINGS = {
    '__': 'Hidden Attributes',
    'NoneType': 'None',
    'type': 'Type',
    'bool': 'Boolean',
    'int': 'Integer',
    'long': 'Long Integer',
    'float': 'Float',
    'complex': 'Complex',
    'str': 'String',
    'unicode': 'Unicode String',
    'tuple': 'Tuple',
    'list': 'List/Array',
    'dict': 'Dictionary/Hash/Map',
    'dict-proxy': 'Dictionary Proxy',
    'set': 'Set',
    'frozenset': 'Frozen Set',
    'file': 'File',
    'xrange': 'X Range',
    'slice': 'Slice',
    'buffer': 'Buffer',
    'class': 'Class',
    'instance': 'Class Instance',
    'method': 'Class Method',
    'property': 'Class Property',
    'generator': 'Generator',
    'function': 'Function',
    'builtin_function_or_method': 'Builtin Function',
    'code': 'Code',
    'module': 'Module',
    'ellipsis': 'Ellipsis',
    'traceback': 'Traceback',
    'frame': 'Frame',
    'other': 'Other'}



# -*- coding: utf-8 -*-
#
# codimension - graphics python two-way code editor and analyzer
# Copyright (C) 2010-2016  Sergey Satskiy <sergey.satskiy@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#


"""codimension project"""

# pylint: disable=W0702
# pylint: disable=W0703

import logging
import uuid
import re
import copy
import json
import shutil
import os
from os.path import (realpath, islink, isdir, sep, exists, dirname, isabs,
                     join, relpath, isfile, basename)
from ui.qt import QObject, pyqtSignal
from .settings import Settings, SETTINGS_DIR
from .watcher import Watcher
from .config import DEFAULT_ENCODING
from .debugenv import DebuggerEnvironment
from .searchenv import SearchEnvironment
from .fsenv import FileSystemEnvironment
from .runparamscache import RunParametersCache
from .filepositions import FilePositions
from .userencodings import FileEncodings
from .flowgroups import FlowUICollapsedGroups


# Saved in .cdm3 file
_DEFAULT_PROJECT_PROPS = {'scriptname': '',    # Script to run the project
                          'mddocfile': '',
                          'creationdate': '',
                          'author': '',
                          'license': '',
                          'copyright': '',
                          'version': '',
                          'email': '',
                          'description': '',
                          'uuid': '',
                          'importdirs': [],
                          'encoding': ''}


class CodimensionProject(QObject,
                         DebuggerEnvironment,
                         SearchEnvironment,
                         FileSystemEnvironment,
                         RunParametersCache,
                         FilePositions,
                         FileEncodings,
                         FlowUICollapsedGroups):

    """Provides codimension project singleton facility"""

    # Constants for the sigProjectChanged signal
    CompleteProject = 0     # It is a completely new project
    Properties = 1          # Project properties were updated

    sigProjectChanged = pyqtSignal(int)
    sigFSChanged = pyqtSignal(list)
    sigRestoreProjectExpandedDirs = pyqtSignal()
    sigProjectAboutToUnload = pyqtSignal()
    sigRecentFilesChanged = pyqtSignal()

    def __init__(self):
        QObject.__init__(self)
        DebuggerEnvironment.__init__(self)
        SearchEnvironment.__init__(self)
        FileSystemEnvironment.__init__(self)
        RunParametersCache.__init__(self)
        FilePositions.__init__(self)
        FileEncodings.__init__(self)
        FlowUICollapsedGroups.__init__(self)

        self.__dirWatcher = None

        # Avoid pylint complains
        self.fileName = ""
        self.userProjectDir = ""    # Directory in ~/.codimension3/uuidNN/
        self.filesList = set()

        self.props = copy.deepcopy(_DEFAULT_PROJECT_PROPS)

        # Precompile the exclude filters for the project files list
        self.__excludeFilter = []
        for flt in Settings()['projectFilesFilters']:
            self.__excludeFilter.append(re.compile(flt))

    def shouldExclude(self, name):
        """Tests if a file must be excluded"""
        if name == '.pylintrc':
            return False
        for excl in self.__excludeFilter:
            if excl.match(name):
                return True
        return False

    def __resetValues(self):
        """Initializes or resets all the project members"""
        # Empty file name means that the project has not been loaded or
        # created. This must be an absolute path.
        self.fileName = ""
        self.userProjectDir = ""

        # Generated having the project dir Full paths are stored.
        # The set holds all files and directories.
        # The dirs end with os.path.sep
        self.filesList = set()

        self.props = copy.deepcopy(_DEFAULT_PROJECT_PROPS)

        RunParametersCache.reset(self)
        DebuggerEnvironment.reset(self)
        SearchEnvironment.reset(self)
        FileSystemEnvironment.reset(self)
        FilePositions.reset(self)
        FileEncodings.reset(self)
        FlowUICollapsedGroups.reset(self)

        # Reset the dir watchers if so
        if self.__dirWatcher is not None:
            del self.__dirWatcher
            self.__dirWatcher = None

    def createNew(self, fileName, props):
        """Creates a new project"""
        # Try to create the user project directory
        projectUuid = str(uuid.uuid1())
        userProjectDir = SETTINGS_DIR + projectUuid + sep
        if not exists(userProjectDir):
            try:
                os.makedirs(userProjectDir)
            except Exception:
                logging.error('Cannot create user project directory: %s. '
                              'Please check the available disk space, '
                              'permissions and re-create the project.',
                              self.userProjectDir)
                raise
        else:
            logging.warning('The user project directory exists! '
                            'The content will be overwritten.')
            self.__removeProjectFiles(userProjectDir)

        # Basic pre-requisites are met. We can reset the current project.
        self.__resetValues()

        self.fileName = fileName
        props['uuid'] = projectUuid
        self.props = props
        self.userProjectDir = userProjectDir

        self.__createProjectFile()  # ~/.codimension3/uuidNN/project

        RunParametersCache.setup(self, self.userProjectDir)
        DebuggerEnvironment.setup(self, self.userProjectDir)
        SearchEnvironment.setup(self, self.userProjectDir)
        FileSystemEnvironment.setup(self, self.userProjectDir)
        FilePositions.setup(self, self.userProjectDir)
        FileEncodings.setup(self, self.userProjectDir)
        FlowUICollapsedGroups.setup(self, self.userProjectDir)

        self.__generateFilesList()

        self.saveProject()

        # Update the watcher
        self.__dirWatcher = Watcher(Settings()['projectFilesFilters'],
                                    self.getProjectDir())
        self.__dirWatcher.sigFSChanged.connect(self.onFSChanged)

        self.sigProjectChanged.emit(self.CompleteProject)

    @staticmethod
    def __removeProjectFiles(userProjectDir):
        """Removes user project files"""
        for root, dirs, files in os.walk(userProjectDir):
            for f in files:
                try:
                    os.unlink(join(root, f))
                except Exception:
                    pass
            for d in dirs:
                try:
                    shutil.rmtree(join(root, d))
                except Exception:
                    pass

    def __createProjectFile(self):
        """Helper function to create the user project file"""
        try:
            with open(self.userProjectDir + 'project', 'w',
                      encoding=DEFAULT_ENCODING) as diskfile:
                diskfile.write(self.fileName)
        except Exception as exc:
            logging.error('Could not create the %s project file: %s',
                          self.userProjectDir, str(exc))

    def saveProject(self):
        """Writes all the settings into the file"""
        if not self.isLoaded():
            return

        # It could be another user project file without write permissions
        skipProjectFile = False
        if exists(self.fileName):
            if not os.access(self.fileName, os.W_OK):
                skipProjectFile = True
        else:
            if not os.access(dirname(self.fileName), os.W_OK):
                skipProjectFile = True

        if not skipProjectFile:
            with open(self.fileName, 'w',
                      encoding=DEFAULT_ENCODING) as diskfile:
                json.dump(self.props, diskfile, indent=4)
        else:
            logging.warning('Skipping updates in %s due to writing permissions',
                            self.fileName)

    def loadProject(self, projectFile):
        """Loads a project from the given file"""
        path = realpath(projectFile)
        if not exists(path):
            raise Exception('Cannot open project file ' + projectFile)
        if not path.endswith('.cdm3'):
            raise Exception('Unexpected project file extension. '
                            'Expected: .cdm3')

        try:
            with open(path, 'r', encoding=DEFAULT_ENCODING) as diskfile:
                props = json.load(diskfile)
        except:
            # Bad error - cannot load project file at all
            raise Exception('Bad project file ' + projectFile)

        self.__resetValues()
        self.fileName = path
        self.props = props

        # Make sure the old projects have the new fields as well
        for key, value in _DEFAULT_PROJECT_PROPS.items():
            if key not in self.props:
                self.props[key] = value

        if self.props['uuid'] == '':
            logging.warning('Project file does not have UUID. '
                            'Re-generate it...')
            self.props['uuid'] = str(uuid.uuid1())
        self.userProjectDir = SETTINGS_DIR + self.props['uuid'] + sep
        if not exists(self.userProjectDir):
            os.makedirs(self.userProjectDir)

        # Read the other config files
        DebuggerEnvironment.setup(self, self.userProjectDir)
        SearchEnvironment.setup(self, self.userProjectDir)
        FileSystemEnvironment.setup(self, self.userProjectDir)
        RunParametersCache.setup(self, self.userProjectDir)
        FilePositions.setup(self, self.userProjectDir)
        FileEncodings.setup(self, self.userProjectDir)
        FlowUICollapsedGroups.setup(self, self.userProjectDir)

        # The project might have been moved...
        self.__createProjectFile()  # ~/.codimension3/uuidNN/project
        self.__generateFilesList()

        # Update the recent list
        Settings().addRecentProject(self.fileName)

        # Setup the new watcher
        self.__dirWatcher = Watcher(Settings()['projectFilesFilters'],
                                    self.getProjectDir())
        self.__dirWatcher.sigFSChanged.connect(self.onFSChanged)

        self.sigProjectChanged.emit(self.CompleteProject)
        self.sigRestoreProjectExpandedDirs.emit()

    def getImportDirsAsAbsolutePaths(self):
        """Provides a list of import dirs as absolute paths"""
        result = []
        for path in self.props['importdirs']:
            if isabs(path):
                result.append(path)
            else:
                result.append(self.getProjectDir() + path)
        return result

    def onFSChanged(self, items):
        """Triggered when the watcher detects changes"""
        for item in items:
            try:
                if item.startswith('+'):
                    self.filesList.add(item[1:])
                else:
                    self.filesList.remove(item[1:])
            except:
                pass
        self.sigFSChanged.emit(items)

    def unloadProject(self, emitSignal=True):
        """Unloads the current project if required"""
        self.sigProjectAboutToUnload.emit()
        self.__resetValues()
        if emitSignal:
            # No need to send a signal e.g. if IDE is closing
            self.sigProjectChanged.emit(self.CompleteProject)

    def setImportDirs(self, paths):
        """Sets a new set of the project import dirs"""
        if self.props['importdirs'] != paths:
            self.props['importdirs'] = paths
            self.saveProject()
            self.sigProjectChanged.emit(self.Properties)

    def __generateFilesList(self):
        """Generates the files list having the list of dirs"""
        self.filesList = set()
        path = self.getProjectDir()
        self.filesList.add(path)
        self.__scanDir(path)

    def __scanDir(self, path):
        """Recursive function to scan one dir"""
        # The path is with '/' at the end
        for item in os.listdir(path):
            if self.shouldExclude(item):
                continue

            # Exclude symlinks if they point to the other project
            # covered pieces
            candidate = path + item
            if islink(candidate):
                realItem = realpath(candidate)
                if isdir(realItem):
                    if self.isProjectDir(realItem):
                        continue
                else:
                    if self.isProjectDir(dirname(realItem)):
                        continue

            if isdir(candidate):
                self.filesList.add(candidate + sep)
                self.__scanDir(candidate + sep)
                continue
            self.filesList.add(candidate)

    def isProjectDir(self, path):
        """Returns True if the path belongs to the project"""
        if not self.isLoaded():
            return False
        path = realpath(path)     # it could be a symlink
        if not path.endswith(sep):
            path += sep
        return path.startswith(self.getProjectDir())

    def isProjectFile(self, path):
        """Returns True if the path belongs to the project"""
        if not self.isLoaded():
            return False
        return self.isProjectDir(dirname(path))

    def isTopLevelDir(self, path):
        """Checks if the path is a top level dir"""
        if not path.endswith(sep):
            path += sep
        return path in self.topLevelDirs

    def updateProperties(self, props):
        """Updates the project properties"""
        if self.props != props:
            self.props = props
            self.saveProject()
            self.sigProjectChanged.emit(self.Properties)

    def onProjectFileUpdated(self):
        """Called when a project file is updated via direct editing"""
        self.props = getProjectProperties(self.fileName)

        # no need to save, but signal just in case
        self.sigProjectChanged.emit(self.Properties)

    def isLoaded(self):
        """Returns True if a project is loaded"""
        return self.fileName != ''

    def getProjectDir(self):
        """Provides an absolute path to the project dir"""
        if not self.isLoaded():
            return None
        return dirname(realpath(self.fileName)) + sep

    def getProjectName(self):
        """Provides the project name or None"""
        if not self.isLoaded():
            return None

        fBaseName = basename(self.fileName)
        if '.' in fBaseName:
            return fBaseName.split('.')[0].strip()
        return fBaseName

    def getProjectScript(self):
        """Provides the project script file name"""
        if not self.isLoaded():
            return None
        if self.props['scriptname'] == '':
            return None
        if isabs(self.props['scriptname']):
            return self.props['scriptname']
        return realpath(self.getProjectDir() + self.props['scriptname'])

    def addRecentFile(self, path):
        """Adds a recent file. True if a new file was inserted."""
        ret = FileSystemEnvironment.addRecentFile(self, path)
        if ret:
            self.sigRecentFilesChanged.emit()
        return ret

    def getRelativePath(self, path):
        """Provides a relative path if so"""
        if self.isProjectFile(path):
            return relpath(path, dirname(self.fileName))
        return path

    def getAbsolutePath(self, path):
        """Provides an absolute path if so"""
        if isabs(path):
            return path
        if self.isLoaded():
            return join(dirname(self.fileName), path)
        return path

    def getStartupMarkdownFile(self):
        """Provides the startup documentation markdown file if so"""
        if not self.isLoaded():
            return None
        # Could be in project properties
        if not self.props['mddocfile']:
            return None
        if isabs(self.props['mddocfile']):
            return self.props['mddocfile']
        return realpath(self.getProjectDir() + self.props['mddocfile'])

    def findStartupMarkdownFile(self):
        """Finds the startup MD doc file"""
        if not self.isLoaded():
            return None, None
        fName = self.getStartupMarkdownFile()
        if fName:
            if not isabs(fName):
                fName = self.getAbsolutePath(fName)
            if not exists(fName):
                return None, 'Configured markdown doc file ' + \
                       self.getStartupMarkdownFile() + ' is not found'
            return fName, None

        # check the file system
        projectDir = self.getProjectDir()
        for item in os.listdir(projectDir):
            if isfile(projectDir + item):
                lowerName = item.lower()
                if lowerName.endswith('.md') and lowerName.startswith('readme'):
                    return projectDir + item, None
        return None, None

    def suggestStartupMarkdownFile(self):
        """Suggests the default project doc file name"""
        if not self.isLoaded():
            raise Exception('Invalid logic. A markdown project doc file name '
                            'is requested without a loaded project')
        return self.getProjectDir() + 'README.md'



def getProjectProperties(projectFile):
    """Provides project properties or throws an exception"""
    path = realpath(projectFile)
    if not exists(path):
        raise Exception("Cannot find project file " + projectFile)

    try:
        with open(path, 'r', encoding=DEFAULT_ENCODING) as diskfile:
            return json.load(diskfile)
    except Exception as exc:
        logging.error('Error reading project file %s: %s',
                      projectFile, str(exc))
        return {}


def getProjectFileTooltip(fileName):
    """Provides a project file tooltip"""
    props = getProjectProperties(fileName)
    return '\n'.join(['Version: ' + props.get('version', 'n/a'),
                      'Description: ' + props.get('description', 'n/a'),
                      'Author: ' + props.get('author', 'n/a'),
                      'e-mail: ' + props.get('email', 'n/a'),
                      'Copyright: ' + props.get('copyright', 'n/a'),
                      'License: ' + props.get('license', 'n/a'),
                      'Creation date: ' + props.get('creationdate', 'n/a'),
                      'Default encoding: ' + props.get('encoding', 'n/a'),
                      'UUID: ' + props.get('uuid', 'n/a')])
